# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from prometheus_client.core import GaugeMetricFamily, CounterMetricFamily, InfoMetricFamily
from mktxp.cli.config.config import MKTXPConfigKeys, config_handler

class BaseCollector:
    """ Base Collector methods
        For use by custom collector
    """
    @staticmethod
    def _de_duplicate_records(router_records, metric_labels, verbose_reporting = None):
        unique_records = {}
        if verbose_reporting is None:
            verbose_reporting = config_handler.system_entry.verbose_mode

        for record in router_records:
            label_values = tuple(record.get(label, '') for label in metric_labels)
            if label_values in unique_records:
                if verbose_reporting: 
                    print(f"Warning: Duplicate metric record found for labels {dict(zip(metric_labels, label_values))}. Keeping last.")
            unique_records[label_values] = record
        return unique_records.values()

    @staticmethod
    def info_collector(name: str, documentation: str, router_records, metric_labels=None, verbose_reporting=None):
        metric_labels = metric_labels or []
        router_records = router_records or []

        BaseCollector._add_id_labels(metric_labels)
        collector = InfoMetricFamily(f'mktxp_{name}', documentation=documentation)

        deduplicated_records = BaseCollector._de_duplicate_records(router_records, metric_labels, verbose_reporting)
        for router_record in deduplicated_records:
            label_values = {label: router_record.get(label) if router_record.get(label) else '' for label in metric_labels}
            collector.add_metric(metric_labels, label_values)
        return collector

    @staticmethod
    def counter_collector(name: str, documentation: str, router_records, metric_key, metric_labels=None, verbose_reporting=None):
        metric_labels = metric_labels or []
        router_records = router_records or []            

        BaseCollector._add_id_labels(metric_labels)
        collector = CounterMetricFamily(f'mktxp_{name}', documentation=documentation, labels=metric_labels)

        deduplicated_records = BaseCollector._de_duplicate_records(router_records, metric_labels, verbose_reporting)
        for router_record in deduplicated_records:
            label_values = [router_record.get(label) if router_record.get(label) else '' for label in metric_labels]
            collector.add_metric(label_values, router_record.get(metric_key, 0))
        return collector

    @staticmethod
    def gauge_collector(name: str, documentation: str, router_records, metric_key, metric_labels = None, add_id_labels = True, verbose_reporting=None):
        metric_labels = metric_labels or []
        router_records = router_records or []            

        if add_id_labels:
            BaseCollector._add_id_labels(metric_labels)
        collector = GaugeMetricFamily(f'mktxp_{name}', documentation=documentation, labels=metric_labels)

        deduplicated_records = BaseCollector._de_duplicate_records(router_records, metric_labels, verbose_reporting)
        for router_record in deduplicated_records:       
            label_values = [router_record.get(label) if router_record.get(label) else '' for label in metric_labels]        
            collector.add_metric(label_values, router_record.get(metric_key, 0))
        return collector

    # Helpers
    @staticmethod
    def _add_id_labels(metric_labels):
        if MKTXPConfigKeys.ROUTERBOARD_NAME not in metric_labels:
            metric_labels.append(MKTXPConfigKeys.ROUTERBOARD_NAME)
        if MKTXPConfigKeys.ROUTERBOARD_ADDRESS not in metric_labels:
            metric_labels.append(MKTXPConfigKeys.ROUTERBOARD_ADDRESS)



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.public_ip_ds import PublicIPAddressDatasource


class PublicIPAddressCollector(BaseCollector):
    '''Public IP address collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.public_ip:
            return

        address_labels = ['public_address', 'public_address_ipv6', 'dns_name']
        address_records = PublicIPAddressDatasource.metric_records(router_entry, metric_labels=address_labels)

        if address_records:
            for address_record in address_records:
                if not 'dns_name' in address_record:
                    address_record['dns_name'] = 'ddns disabled'

            address_metrics = BaseCollector.info_collector('public_ip_address', 'Public IP address', address_records, address_labels)
            yield address_metrics



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class MKTXPMetricsDataSource:
    ''' MKTXP Metrics data provider
    '''             
    @staticmethod
    def metric_records(router_entry):
        mktxp_records = []
        for key in router_entry.time_spent.keys():
            mktxp_records.append({'name': key, 'duration': router_entry.time_spent[key]})           

        # translation rules            
        translation_table = {'duration': lambda d: d*1000}
        return BaseDSProcessor.trimmed_records(router_entry, router_records = mktxp_records, translation_table = translation_table)



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class POEMetricsDataSource:
    ''' POE Metrics data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, include_comments = False, metric_labels = None):
        if metric_labels is None:
            metric_labels = []                
        try:
            poe_records = router_entry.api_connection.router_api().get_resource('/interface/ethernet/poe').get()
            for int_num, poe_record in enumerate(poe_records):
                poe_monitor_records = router_entry.api_connection.router_api().get_resource('/interface/ethernet/poe').call('monitor', {'once':'', 'numbers':f'{int_num}'})
                poe_monitor_records = BaseDSProcessor.trimmed_records(router_entry, router_records = poe_monitor_records)

                if poe_monitor_records[0].get('poe_out_status'):
                    poe_record['poe_out_status'] = poe_monitor_records[0]['poe_out_status']

                if poe_monitor_records[0].get('poe_out_voltage'):
                    poe_record['poe_out_voltage'] = poe_monitor_records[0]['poe_out_voltage']

                if poe_monitor_records[0].get('poe_out_current'):
                    poe_record['poe_out_current'] = poe_monitor_records[0]['poe_out_current']

                if poe_monitor_records[0].get('poe_out_power'):
                    poe_record['poe_out_power'] = poe_monitor_records[0]['poe_out_power']

            if include_comments:
                interfaces = router_entry.api_connection.router_api().get_resource('/interface/ethernet').call('print', {'proplist':'name,comment'})
                comment_fn = lambda interface: interface['comment'] if interface.get('comment') else ''
                for poe_record in poe_records:
                    comment = [comment_fn(interface) for interface in interfaces if interface['name'] == poe_record['name']][0]       
                    if comment:
                        # combines name with comment
                        poe_record['name'] = comment if router_entry.config_entry.use_comments_over_names else \
                                                                                                    f"{poe_record['name']} ({comment})"                                
            return BaseDSProcessor.trimmed_records(router_entry, router_records = poe_records, metric_labels = metric_labels)
        except Exception as exc:
            print(f'Error getting PoE info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.ipsec_ds import IPSecMetricsDataSource


class IPSecCollector(BaseCollector):
    '''IPSec collector'''

    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.ipsec:
            return

        ipsec_labels = ['local_address', 'remote_address', 'name', 'last_seen', 'uptime', 'ph2_total', 'responder',
                        'natt_peer', 'rx_bytes', 'rx_packets', 'tx_bytes', 'tx_packets', 'state']
        translation_table = {
            'responder': lambda value: '1' if value == 'true' else '0',
            'natt_peer': lambda value: '1' if value == 'true' else '0',
            'last_seen': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value) if value else '0',
            'uptime': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value) if value else '0'
        }
        ipsec_records = IPSecMetricsDataSource.metric_records(router_entry, metric_labels=ipsec_labels,
                                                              translation_table=translation_table)

        if ipsec_records:
            ipsec_info_labels = ['local_address', 'name', 'remote_address', 'state']
            yield BaseCollector.info_collector('ipsec_peer_state',
                                               'State of negotiation with the peer.',
                                               ipsec_records, ipsec_info_labels)

            ipsec_value_labels = ['local_address', 'name', 'remote_address']
            yield BaseCollector.counter_collector('ipsec_peer_rx_byte',
                                                  'The total amount of bytes received from this peer.',
                                                  ipsec_records, 'rx_bytes', ipsec_value_labels)

            yield BaseCollector.counter_collector('ipsec_peer_tx_byte',
                                                  'The total amount of bytes transmitted to this peer.',
                                                  ipsec_records, 'tx_bytes', ipsec_value_labels)

            yield BaseCollector.counter_collector('ipsec_peer_rx_packet',
                                                  'The total amount of packets received from this peer.',
                                                  ipsec_records, 'rx_packets', ipsec_value_labels)

            yield BaseCollector.counter_collector('ipsec_peer_tx_packet',
                                                  'The total amount of packets transmitted to this peer.',
                                                  ipsec_records, 'tx_packets', ipsec_value_labels)

            yield BaseCollector.gauge_collector('ipsec_peer_security_association',
                                                'The total amount of active IPsec security associations.',
                                                ipsec_records, 'ph2_total', ipsec_value_labels)

            yield BaseCollector.gauge_collector('ipsec_peer_last_seen',
                                                'Duration since the last message received by this peer.',
                                                ipsec_records, 'last_seen', ipsec_value_labels)

            yield BaseCollector.gauge_collector('ipsec_peer_uptime',
                                                'How long peer is in an established state.',
                                                ipsec_records, 'uptime', ipsec_value_labels)

            yield BaseCollector.gauge_collector('ipsec_peer_responder',
                                                'Whether the connection is initiated by a remote peer.',
                                                ipsec_records, 'responder', ipsec_value_labels)

            yield BaseCollector.gauge_collector('ipsec_peer_natt_enabled',
                                                'Whether NAT-T is used for this peer.',
                                                ipsec_records, 'natt_peer', ipsec_value_labels)



# coding=utf8
# Copyright (c) 2020 Arseniy Kuznetsov
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


from mktxp.cli.config.config import MKTXPConfigKeys
from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.dns_ds import DNSDataSource


class DNSCollector(BaseCollector):
    '''Dns Collector'''

    @staticmethod
    def collect(router_entry):
        allowed_properties = {'cache_size', 'cache_used'}
        if not router_entry.config_entry.dns:
            return

        record = DNSDataSource.metric_records(router_entry)

        if not record:
            return

        keys = list(record.keys())
        metrics = []

        for key in keys:
            if key not in allowed_properties:
                continue

            metric_record = {
                MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                'property': key,
                'value': int(record[key]) * 1024
            }
            metrics.append(metric_record)

        yield BaseCollector.gauge_collector(
            'dns_info',
            'DNS info',
            metrics,
            'value',
            metric_labels=['property']
        )



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from collections import OrderedDict
from mktxp.cli.config.config import CollectorKeys
from mktxp.collector.dhcp_collector import DHCPCollector
from mktxp.collector.package_collector import PackageCollector
from mktxp.collector.connection_collector import IPConnectionCollector
from mktxp.collector.interface_collector import InterfaceCollector
from mktxp.collector.health_collector import HealthCollector
from mktxp.collector.identity_collector import IdentityCollector
from mktxp.collector.ipsec_collector import IPSecCollector
from mktxp.collector.public_ip_collector import PublicIPAddressCollector
from mktxp.collector.neighbor_collector import NeighborCollector
from mktxp.collector.monitor_collector import MonitorCollector
from mktxp.collector.poe_collector import POECollector
from mktxp.collector.netwatch_collector import NetwatchCollector
from mktxp.collector.pool_collector import PoolCollector
from mktxp.collector.resource_collector import SystemResourceCollector
from mktxp.collector.route_collector import RouteCollector
from mktxp.collector.wlan_collector import WLANCollector
from mktxp.collector.capsman_collector import CapsmanCollector
from mktxp.collector.bandwidth_collector import BandwidthCollector
from mktxp.collector.firewall_collector import FirewallCollector
from mktxp.collector.mktxp_collector import MKTXPCollector
from mktxp.collector.user_collector import UserCollector
from mktxp.collector.queue_collector import QueueTreeCollector
from mktxp.collector.queue_collector import QueueSimpleCollector
from mktxp.collector.kid_control_device_collector import KidDeviceCollector
from mktxp.collector.bfd_collector import BFDCollector
from mktxp.collector.bgp_collector import BGPCollector
from mktxp.collector.routing_stats_collector import RoutingStatsCollector
from mktxp.collector.eoip_collector import EOIPCollector
from mktxp.collector.gre_collector import GRECollector
from mktxp.collector.ipip_collector import IPIPCollector
from mktxp.collector.lte_collector import LTECollector
from mktxp.collector.switch_collector import SwitchPortCollector
from mktxp.collector.certificate_collector import CertificateCollector
from mktxp.collector.dns_collector import DNSCollector
from mktxp.collector.container_collector import ContainerCollector

class CollectorRegistry:
    ''' MKTXP Collectors Registry
    '''
    def __init__(self):
        self.registered_collectors = OrderedDict()

        # bandwidth collector is not router-entry related, so registering directly
        self.bandwidthCollector = BandwidthCollector()

        self.register(CollectorKeys.IDENTITY_COLLECTOR, IdentityCollector.collect)
        self.register(CollectorKeys.SYSTEM_RESOURCE_COLLECTOR, SystemResourceCollector.collect)
        self.register(CollectorKeys.HEALTH_COLLECTOR, HealthCollector.collect)
        self.register(CollectorKeys.PUBLIC_IP_ADDRESS_COLLECTOR, PublicIPAddressCollector.collect)

        self.register(CollectorKeys.NEIGHBOR_COLLECTOR, NeighborCollector.collect)
        self.register(CollectorKeys.DNS_COLLECTOR, DNSCollector.collect)

        self.register(CollectorKeys.PACKAGE_COLLECTOR, PackageCollector.collect)
        self.register(CollectorKeys.DHCP_COLLECTOR, DHCPCollector.collect)
        self.register(CollectorKeys.IP_CONNECTION_COLLECTOR, IPConnectionCollector.collect)
        self.register(CollectorKeys.IPSEC_COLLECTOR, IPSecCollector.collect)
        self.register(CollectorKeys.POOL_COLLECTOR, PoolCollector.collect)
        self.register(CollectorKeys.INTERFACE_COLLECTOR, InterfaceCollector.collect)

        self.register(CollectorKeys.FIREWALL_COLLECTOR, FirewallCollector.collect)
        self.register(CollectorKeys.MONITOR_COLLECTOR, MonitorCollector.collect)
        self.register(CollectorKeys.POE_COLLECTOR, POECollector.collect)
        self.register(CollectorKeys.NETWATCH_COLLECTOR, NetwatchCollector.collect)
        self.register(CollectorKeys.ROUTE_COLLECTOR, RouteCollector.collect)

        self.register(CollectorKeys.WLAN_COLLECTOR, WLANCollector.collect)
        self.register(CollectorKeys.CAPSMAN_COLLECTOR, CapsmanCollector.collect)

        self.register(CollectorKeys.USER_COLLECTOR, UserCollector.collect)
        self.register(CollectorKeys.QUEUE_TREE_COLLECTOR, QueueTreeCollector.collect)
        self.register(CollectorKeys.QUEUE_SIMPLE_COLLECTOR, QueueSimpleCollector.collect)

        self.register(CollectorKeys.KID_CONTROL_DEVICE_COLLECTOR, KidDeviceCollector.collect)
        self.register(CollectorKeys.BFD_COLLECTOR, BFDCollector.collect)
        self.register(CollectorKeys.BGP_COLLECTOR, BGPCollector.collect)
        self.register(CollectorKeys.EOIP_COLLECTOR, EOIPCollector.collect)
        self.register(CollectorKeys.GRE_COLLECTOR, GRECollector.collect)
        self.register(CollectorKeys.IPIP_COLLECTOR, IPIPCollector.collect)
        self.register(CollectorKeys.LTE_COLLECTOR, LTECollector.collect)
        self.register(CollectorKeys.SWITCH_PORT_COLLECTOR, SwitchPortCollector.collect)

        self.register(CollectorKeys.CONTAINER_COLLECTOR, ContainerCollector.collect)

        self.register(CollectorKeys.ROUTING_STATS_COLLECTOR, RoutingStatsCollector.collect)
        self.register(CollectorKeys.CERTIFICATE_COLLECTOR, CertificateCollector.collect)

        self.register(CollectorKeys.MKTXP_COLLECTOR, MKTXPCollector.collect)
        

    def register(self, collector_ID, collect_func):
        self.registered_collectors[collector_ID] = collect_func





# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from enum import IntEnum
from collections import namedtuple
from mktxp.cli.config.config import config_handler, MKTXPConfigKeys, CollectorKeys
from mktxp.flow.router_connection import RouterAPIConnection
from mktxp.datasource.package_ds import PackageMetricsDataSource
from mktxp.datasource.system_resource_ds import SystemResourceMetricsDataSource
from mktxp.flow.router_connection import RouterAPIConnectionError

class RouterEntryWirelessType(IntEnum):
    NONE = 0
    WIRELESS = 1
    WIFIWAVE2 = 2
    WIFI = 3
    DUAL = 4

class RouterEntryWirelessPackage:
    WIFI_PACKAGE = 'wifi-qcom'
    WIFI_AC_PACKAGE = 'wifi-qcom-ac'
    WIFIWAVE2_PACKAGE = 'wifiwave2'
    WIRELESS_PACKAGE = 'wireless'

class RouterEntryConnectionState(IntEnum):
    NOT_CONNECTED = 0
    PARTIALLY_CONNECTED = 1
    CONNECTED = 2

class RouterEntry:
    ''' RouterOS Entry
    '''                 
    def __init__(self, router_name):
        self.router_name = router_name
        self.config_entry = config_handler.config_entry(router_name)
        self.api_connection = RouterAPIConnection(router_name, self.config_entry)
        self.router_id = {
            MKTXPConfigKeys.ROUTERBOARD_NAME: self.router_name,
            MKTXPConfigKeys.ROUTERBOARD_ADDRESS: self.config_entry.hostname
            }
        
        self.time_spent =  { CollectorKeys.IDENTITY_COLLECTOR: 0,
                            CollectorKeys.SYSTEM_RESOURCE_COLLECTOR: 0,
                            CollectorKeys.HEALTH_COLLECTOR: 0,
                            CollectorKeys.PUBLIC_IP_ADDRESS_COLLECTOR: 0,
                            CollectorKeys.NEIGHBOR_COLLECTOR: 0,
                            CollectorKeys.DNS_COLLECTOR: 0,
                            CollectorKeys.PACKAGE_COLLECTOR: 0,
                            CollectorKeys.DHCP_COLLECTOR: 0,
                            CollectorKeys.POOL_COLLECTOR: 0,
                            CollectorKeys.IP_CONNECTION_COLLECTOR: 0,
                            CollectorKeys.IPSEC_COLLECTOR: 0,
                            CollectorKeys.INTERFACE_COLLECTOR: 0,
                            CollectorKeys.FIREWALL_COLLECTOR: 0,
                            CollectorKeys.MONITOR_COLLECTOR: 0,
                            CollectorKeys.POE_COLLECTOR: 0,
                            CollectorKeys.NETWATCH_COLLECTOR: 0,
                            CollectorKeys.ROUTE_COLLECTOR: 0,
                            CollectorKeys.WLAN_COLLECTOR: 0,
                            CollectorKeys.CAPSMAN_COLLECTOR: 0,
                            CollectorKeys.QUEUE_TREE_COLLECTOR: 0,
                            CollectorKeys.QUEUE_SIMPLE_COLLECTOR: 0,                            
                            CollectorKeys.KID_CONTROL_DEVICE_COLLECTOR: 0,
                            CollectorKeys.USER_COLLECTOR: 0,
                            CollectorKeys.BFD_COLLECTOR: 0,
                            CollectorKeys.BGP_COLLECTOR: 0,
                            CollectorKeys.ROUTING_STATS_COLLECTOR: 0,
                            CollectorKeys.EOIP_COLLECTOR: 0,
                            CollectorKeys.GRE_COLLECTOR: 0,
                            CollectorKeys.IPIP_COLLECTOR: 0,
                            CollectorKeys.LTE_COLLECTOR: 0,
                            CollectorKeys.SWITCH_PORT_COLLECTOR: 0,
                            CollectorKeys.MKTXP_COLLECTOR: 0,
                            CollectorKeys.CERTIFICATE_COLLECTOR: 0,
                            CollectorKeys.CONTAINER_COLLECTOR: 0
                            }         
        self._dhcp_entry = None        
        self._dhcp_records = {}
        self._capsman_entry = None
        self._wireless_type = RouterEntryWirelessType.NONE                                    

    @property
    def wireless_type(self):
        router_entry = self
        if self._wireless_type == RouterEntryWirelessType.NONE:
            if PackageMetricsDataSource.is_package_installed(router_entry, package_name = RouterEntryWirelessPackage.WIFI_PACKAGE):
              self._wireless_type = RouterEntryWirelessType.WIFI
            elif PackageMetricsDataSource.is_package_installed(router_entry, package_name = RouterEntryWirelessPackage.WIFI_AC_PACKAGE):
              self._wireless_type = RouterEntryWirelessType.WIFI
            elif PackageMetricsDataSource.is_package_installed(router_entry, package_name = RouterEntryWirelessPackage.WIFIWAVE2_PACKAGE):
              self._wireless_type = RouterEntryWirelessType.WIFIWAVE2
            elif PackageMetricsDataSource.is_package_installed(router_entry, package_name = RouterEntryWirelessPackage.WIRELESS_PACKAGE):
              self._wireless_type = RouterEntryWirelessType.DUAL
            elif SystemResourceMetricsDataSource.has_builtin_wifi_capsman(router_entry):
              self._wireless_type = RouterEntryWirelessType.WIFI
            else:
              self._wireless_type = RouterEntryWirelessType.WIRELESS
        return self._wireless_type

    @property
    def dhcp_entry(self):
        if self._dhcp_entry:
            return self._dhcp_entry
        return self
    @dhcp_entry.setter
    def dhcp_entry(self, dhcp_entry):
        self._dhcp_entry = dhcp_entry

    @property
    def capsman_entry(self):
        if self._capsman_entry:
            return self._capsman_entry
        return self
    @capsman_entry.setter
    def capsman_entry(self, capsman_entry):
        self._capsman_entry = capsman_entry

    @property
    def dhcp_records(self):
        return (entry.record for key, entry in  self._dhcp_records.items() if entry.type == 'mac_address') \
                                                                                if self._dhcp_records else None   
    @dhcp_records.setter
    def dhcp_records(self, dhcp_records):
        for dhcp_record in dhcp_records:
            if dhcp_record.get('mac_address'):
                self._dhcp_records[dhcp_record.get('mac_address')] = DHCPCacheEntry('mac_address', dhcp_record)
            if dhcp_record.get('address'):
                dhcp_record['type'] = 'address'
                self._dhcp_records[dhcp_record.get('address')] = DHCPCacheEntry('address', dhcp_record)

    def dhcp_record(self, key):
        if self._dhcp_records and self._dhcp_records.get(key):
            return self._dhcp_records[key].record
        return None

    def connection_status(self):
        primary_connection_status = self.api_connection.is_connected()
        dhcp_connection_status = self.dhcp_entry.api_connection.is_connected()
        capsman_connection_status = self.capsman_entry.api_connection.is_connected()

        if primary_connection_status and dhcp_connection_status and capsman_connection_status:
            return RouterEntryConnectionState.CONNECTED

        if primary_connection_status or dhcp_connection_status or capsman_connection_status:
            return RouterEntryConnectionState.PARTIALLY_CONNECTED

        return RouterEntryConnectionState.NOT_CONNECTED

    def connect(self):
        if not self.api_connection.is_connected():
            try:
                self.api_connection.connect()
            except RouterAPIConnectionError as exc:
                print (f'{exc}')
        
        if self._dhcp_entry and not self._dhcp_entry.api_connection.is_connected():            
            try:
                self._dhcp_entry.api_connection.connect()
            except RouterAPIConnectionError as exc:
                print (f'{exc}')

        if self._capsman_entry and not self._capsman_entry.api_connection.is_connected():
            try:
                self._capsman_entry.api_connection.connect()
            except RouterAPIConnectionError as exc:
                print (f'{exc}')

    def is_ready(self):           
        # self.is_done() #flush caches, just in case
        is_ready = False

        if self.connection_status() in (RouterEntryConnectionState.NOT_CONNECTED, RouterEntryConnectionState.PARTIALLY_CONNECTED):
            self.connect()
        if self.connection_status() in (RouterEntryConnectionState.CONNECTED, RouterEntryConnectionState.PARTIALLY_CONNECTED):            
            is_ready = True        
        
        return is_ready

    def is_done(self):
        if not config_handler.system_entry.persistent_router_connection_pool:
            self.api_connection.disconnect()
            if self._dhcp_entry:
                self._dhcp_entry.api_connection.disconnect()
            if self._capsman_entry:
                self._capsman_entry.api_connection.disconnect()

        self._dhcp_records = {}
        self._wireless_type = RouterEntryWirelessType.NONE

DHCPCacheEntry = namedtuple('DHCPCacheEntry', ['type', 'record'])



# coding=utf8
# Copyright (c) 2020 Arseniy Kuznetsov
##
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
##
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.neighbor_ds import NeighborDataSource


class NeighborCollector(BaseCollector):
    '''Neighbor Collector'''

    @staticmethod
    def collect(router_entry):
        if router_entry.config_entry.neighbor:
            metric_labels = ['address', 'interface', 'mac_address', 'identity']
            records = NeighborDataSource.metric_records(router_entry, metric_labels=metric_labels)
            metrics = BaseCollector.info_collector('neighbor', 'Reachable neighbors (IPv4)', records, metric_labels=metric_labels)
            yield metrics

        if router_entry.config_entry.ipv6_neighbor:
            metric_labels = ['address', 'interface', 'mac_address', 'status', 'comment']
            records = NeighborDataSource.metric_records(router_entry, metric_labels=metric_labels)
            metrics = BaseCollector.info_collector('ipv6_neighbor', 'Reachable neighbors (IPv6)', records, metric_labels=metric_labels)
            yield metrics



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor
from mktxp.utils.utils import parse_mkt_uptime


class DHCPMetricsDataSource:
    ''' DHCP Metrics data provider
    '''
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None, add_router_id = True, dhcp_cache = True, translate = True, bound = False):
        if metric_labels is None or dhcp_cache:
            metric_labels = ['host_name', 'comment', 'active_address', 'address', 'mac_address', 'server', 'expires_after', 'client_id', 'active_mac_address']

        if dhcp_cache and router_entry.dhcp_records:
            return router_entry.dhcp_records
        try:
            if bound:
                dhcp_lease_records = router_entry.dhcp_entry.api_connection.router_api().get_resource('/ip/dhcp-server/lease').get(status='bound')
            else:
                dhcp_lease_records = router_entry.dhcp_entry.api_connection.router_api().get_resource('/ip/dhcp-server/lease').call('print', {'active':''})

            # translation rules
            translation_table = {}
            if 'comment' in metric_labels:
                translation_table['comment'] = lambda c: c if c else ''
            if 'host_name' in metric_labels:
                translation_table['host_name'] = lambda c: c if c else ''
            if 'expires_after' in metric_labels and translate:
                translation_table['expires_after'] = lambda c: parse_mkt_uptime(c) if c else 0
            if 'active_address' in metric_labels:
                translation_table['active_address'] = lambda c: c if c else ''

            records = BaseDSProcessor.trimmed_records(router_entry, router_records = dhcp_lease_records, metric_labels = metric_labels, add_router_id = add_router_id, translation_table = translation_table)
            if dhcp_cache:
                router_entry.dhcp_records = records
            return records

        except Exception as exc:
            print(f'Error getting dhcp info from router {router_entry.dhcp_entry.router_name}@{router_entry.dhcp_entry.config_entry.hostname}: {exc}')
            return None



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from mktxp.datasource.base_ds import BaseDSProcessor
from mktxp.datasource.package_ds import PackageMetricsDataSource
from mktxp.flow.router_entry import RouterEntryWirelessType

class WirelessMetricsDataSource:
    ''' Wireless Metrics data provider
    '''             
    WIRELESS = 'wireless'
    WIFIWAVE2 = 'wifiwave2'
    WIFI = 'wifi'

    @staticmethod
    def metric_records(router_entry, *, metric_labels = None, add_router_id = True):
        if metric_labels is None:
            metric_labels = []                
        try:
            wireless_package = WirelessMetricsDataSource.wireless_package(router_entry)
            registration_table_records = router_entry.api_connection.router_api().get_resource(f'/interface/{wireless_package}/registration-table').get()

            # With wifiwave2, Mikrotik renamed the field 'signal-strength' to 'signal' 
            # For backward compatibility, including both variants
            for record in registration_table_records:
                if 'signal' in record:
                    record['signal-strength'] = record['signal']

            return BaseDSProcessor.trimmed_records(router_entry, router_records = registration_table_records, metric_labels = metric_labels, add_router_id = add_router_id,)
        except Exception as exc:
            print(f'Error getting wireless registration table info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None

    @staticmethod
    def wireless_package(router_entry):
        if router_entry.wireless_type in (RouterEntryWirelessType.DUAL, RouterEntryWirelessType.WIRELESS):
            return WirelessMetricsDataSource.WIRELESS
        elif router_entry.wireless_type == RouterEntryWirelessType.WIFIWAVE2:
            return WirelessMetricsDataSource.WIFIWAVE2
        else:
            return WirelessMetricsDataSource.WIFI





# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


class BaseDSProcessor:
    ''' Base Metrics DataSource processing
    '''             

    @staticmethod
    def trimmed_records(router_entry, *, router_records = None, metric_labels = None, add_router_id = True, translation_table = None, translate_if_no_value = True):
        metric_labels = metric_labels or []
        router_records = router_records or []            
        translation_table = translation_table or {}         

        if len(metric_labels) == 0 and len(router_records) > 0:
            metric_labels = [BaseDSProcessor._normalise_keys(key) for key in router_records[0].keys()]
        metric_labels = set(metric_labels)      

        labeled_records = []
        for router_record in router_records:
            translated_record = {BaseDSProcessor._normalise_keys(key): value for (key, value) in router_record.items() if BaseDSProcessor._normalise_keys(key) in metric_labels}

            if add_router_id:
                for key, value in router_entry.router_id.items():
                    translated_record[key] = value
            
            # translate fields if needed
            for key, func in translation_table.items():
                if translate_if_no_value or translated_record.get(key) is not None:
                    translated_record[key] = func(translated_record.get(key))
            labeled_records.append(translated_record)            
        return labeled_records


    @staticmethod
    def _normalise_keys(key):
        chars = ".-"
        for chr in chars:
            if chr in key:
                key = key.replace(chr, "_")     
        return key




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from datetime import datetime
from prometheus_client.core import REGISTRY
from prometheus_client import make_wsgi_app

from mktxp.cli.config.config import config_handler
from mktxp.flow.collector_handler import CollectorHandler
from mktxp.flow.collector_registry import CollectorRegistry
from mktxp.flow.router_entries_handler import RouterEntriesHandler

from mktxp.cli.output.capsman_out import CapsmanOutput
from mktxp.cli.output.wifi_out import WirelessOutput
from mktxp.cli.output.dhcp_out import DHCPOutput
from mktxp.cli.output.conn_stats_out import ConnectionsStatsOutput

import gzip
from waitress import serve

class PrometheusHeadersDeduplicatingMiddleware:
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        captured_status = []
        captured_headers = []

        def custom_start_response(status, headers, exc_info=None):
            captured_status.append(status)
            captured_headers.append(headers)

        app_iter = self.app(environ, custom_start_response)
        body = b''.join(app_iter)

        response_headers = dict(captured_headers[0])
        is_gzipped = response_headers.get('Content-Encoding') == 'gzip'

        if is_gzipped:
            try:
                uncompressed_body = gzip.decompress(body)
            except gzip.BadGzipFile:
                uncompressed_body = body
        else:
            uncompressed_body = body

        lines = uncompressed_body.decode('utf-8').split('\n')
        seen_headers = set()
        output_lines = []
        for line in lines:
            if line.startswith('# HELP') or line.startswith('# TYPE'):
                if line in seen_headers:
                    continue
                seen_headers.add(line)
            output_lines.append(line)

        processed_text = '\n'.join(output_lines)

        if is_gzipped:
            output_body = gzip.compress(processed_text.encode('utf-8'))
        else:
            output_body = processed_text.encode('utf-8')

        status = captured_status[0]
        original_headers = captured_headers[0]

        new_headers = []
        for k, v in original_headers:
            if k.lower() == 'content-length':
                new_headers.append((k, str(len(output_body))))
            else:
                new_headers.append((k, v))

        start_response(status, new_headers)
        return [output_body]

class ExportProcessor:
    ''' Base Export Processing
    '''    
    @staticmethod
    def start():
        REGISTRY.register(CollectorHandler(RouterEntriesHandler(), CollectorRegistry()))
        current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print(f'{current_time} Running HTTP metrics server on: {config_handler.system_entry.listen}')

        app = make_wsgi_app()
        if config_handler.system_entry.prometheus_headers_deduplication:
            if config_handler.system_entry.verbose_mode:
                print(f"Prometheus HELP / TYPE headers de-deplucation is On")
            middleware = PrometheusHeadersDeduplicatingMiddleware(app)
            serve(middleware, listen = config_handler.system_entry.listen)
        else:
            serve(app, listen = config_handler.system_entry.listen)

class OutputProcessor:
    ''' Base CLI Processing
    '''    
    @staticmethod
    def capsman_clients(entry_name):
        router_entry = RouterEntriesHandler.router_entry(entry_name)
        if router_entry:
            CapsmanOutput.clients_summary(router_entry)
        
    @staticmethod
    def wifi_clients(entry_name):
        router_entry = RouterEntriesHandler.router_entry(entry_name)
        if router_entry:
            WirelessOutput.clients_summary(router_entry)
        
    @staticmethod
    def dhcp_clients(entry_name):
        router_entry = RouterEntriesHandler.router_entry(entry_name)
        if router_entry:
            DHCPOutput.clients_summary(router_entry)

    @staticmethod
    def conn_stats(entry_name):
        router_entry = RouterEntriesHandler.router_entry(entry_name)
        if router_entry:
            ConnectionsStatsOutput.clients_summary(router_entry)

            



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.routing_stats_ds import RoutingStatsMetricsDataSource
from mktxp.utils.utils import parse_mkt_uptime


class RoutingStatsCollector(BaseCollector):
    '''Routing Stats collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.routing_stats:
            return

        routing_stats_labels = ['tasks', 'id', 'private_mem_blocks', 'shared_mem_blocks', 'kernel_time', 'process_time', 'max_busy', 'max_calc']
        translation_table = {
                'kernel_time': lambda value:  parse_mkt_uptime(value),
                'process_time': lambda value:  parse_mkt_uptime(value),
                'max_busy': lambda value:  parse_mkt_uptime(value),
                'max_calc': lambda value:  parse_mkt_uptime(value),
            }
        routing_stats_records = RoutingStatsMetricsDataSource.metric_records(router_entry, metric_labels=routing_stats_labels, translation_table = translation_table)
                
        if routing_stats_records:
            session_info_labels = ['tasks', 'id', 'pid']
            routing_stats_processes_metrics = BaseCollector.info_collector('routing_stats_processes', 'Routing Process Stats', routing_stats_records, session_info_labels)
            yield routing_stats_processes_metrics

            session_id_labels = ['tasks']
            routing_stats_private_mem_metrics = BaseCollector.gauge_collector('routing_stats_private_mem', 'Private Memory Blocks Used', routing_stats_records, 'private_mem_blocks', session_id_labels)
            yield routing_stats_private_mem_metrics

            routing_stats_shared_mem_metrics = BaseCollector.gauge_collector('routing_stats_shared_mem', 'Shared Memory Blocks Used', routing_stats_records, 'shared_mem_blocks', session_id_labels)
            yield routing_stats_shared_mem_metrics
            
            kernel_time_metrics = BaseCollector.counter_collector('routing_stats_kernel_time', 'Process Kernel Time', routing_stats_records, 'kernel_time', session_id_labels)
            yield kernel_time_metrics

            process_time_metrics = BaseCollector.counter_collector('routing_stats_process_time', 'Process Time', routing_stats_records, 'process_time', session_id_labels)
            yield process_time_metrics

            max_busy_metrics = BaseCollector.counter_collector('routing_stats_max_busy', 'Max Busy Time', routing_stats_records, 'max_busy', session_id_labels)
            yield max_busy_metrics
            
            max_calc_metrics = BaseCollector.counter_collector('routing_stats_max_calc', 'Max Calc Time', routing_stats_records, 'max_calc', session_id_labels)
            yield max_calc_metrics



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class CertificateMetricsDataSource:
    '''Certificates Metrics data provider
    '''             
    @staticmethod    
    def metric_records(router_entry, *, metric_labels = None, translation_table=None):
        if metric_labels is None:
            metric_labels = []                
        try:
            certificates_records = router_entry.api_connection.router_api().get_resource('/certificate').call('print', {'detail':''})
            return BaseDSProcessor.trimmed_records(router_entry, router_records = certificates_records, 
                                                            metric_labels = metric_labels, translation_table=translation_table)
        except Exception as exc:
            print(f'Error getting certificates info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None


# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.cli.config.config import MKTXPConfigKeys
from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.firewall_ds import FirewallMetricsDataSource


class FirewallCollector(BaseCollector):
    ''' Firewall rules traffic metrics collector
    '''    
    @staticmethod
    def collect(router_entry):
        # Initialize all pool counts, including those currently not used
        # These are the same for both IPv4 and IPv6
        firewall_labels = ['chain', 'action', 'bytes', 'comment', 'log', 'out_interface', 'protocol']

        if router_entry.config_entry.firewall:
            # ~*~*~*~*~*~ IPv4 ~*~*~*~*~*~
            firewall_filter_records = FirewallMetricsDataSource.metric_records(router_entry, metric_labels = firewall_labels, filter_path='filter')
            if firewall_filter_records:           
                metrics_records = [FirewallCollector.metric_record(router_entry, record) for record in firewall_filter_records]
                firewall_filter_metrics = BaseCollector.counter_collector('firewall_filter', 'Total amount of bytes matched by firewall rules', metrics_records, 'bytes', ['name', 'log'])
                yield firewall_filter_metrics

            firewall_raw_records = FirewallMetricsDataSource.metric_records(router_entry, metric_labels = firewall_labels, filter_path='raw')
            if firewall_raw_records:      
                metrics_records = [FirewallCollector.metric_record(router_entry, record) for record in firewall_raw_records]     
                firewall_raw_metrics = BaseCollector.counter_collector('firewall_raw', 'Total amount of bytes matched by raw firewall rules', metrics_records, 'bytes', ['name', 'log'])
                yield firewall_raw_metrics

            filter_nat_records = FirewallMetricsDataSource.metric_records(router_entry, metric_labels = firewall_labels, filter_path='nat')
            if filter_nat_records:
                metrics_records = [FirewallCollector.metric_record(router_entry, record) for record in filter_nat_records]
                filter_nat_metrics = BaseCollector.counter_collector('firewall_nat', 'Total amount of bytes matched by NAT rules', metrics_records, 'bytes', ['name', 'log'])
                yield filter_nat_metrics

            filter_mangle_records = FirewallMetricsDataSource.metric_records(router_entry, metric_labels = firewall_labels, filter_path='mangle')
            if filter_mangle_records:
                metrics_records = [FirewallCollector.metric_record(router_entry, record) for record in filter_mangle_records]
                filter_mangle_metrics = BaseCollector.counter_collector('firewall_mangle', 'Total amount of bytes matched by Mangle rules', metrics_records, 'bytes', ['name', 'log'])
                yield filter_mangle_metrics

        # ~*~*~*~*~*~ IPv6 ~*~*~*~*~*~
        if router_entry.config_entry.ipv6_firewall:
            firewall_filter_records_ipv6 =  FirewallMetricsDataSource.metric_records(router_entry, metric_labels = firewall_labels, filter_path='filter', ipv6=True)
            if firewall_filter_records_ipv6:           
                metrics_records_ipv6 = [FirewallCollector.metric_record(router_entry, record) for record in firewall_filter_records_ipv6]
                firewall_filter_metrics_ipv6 = BaseCollector.counter_collector('firewall_filter_ipv6', 'Total amount of bytes matched by firewall rules (IPv6)', metrics_records_ipv6, 'bytes', ['name', 'log'])
                yield firewall_filter_metrics_ipv6

            firewall_raw_records_ipv6 = FirewallMetricsDataSource.metric_records(router_entry, metric_labels = firewall_labels, filter_path='raw', ipv6=True)
            if firewall_raw_records_ipv6:      
                metrics_records_ipv6 = [FirewallCollector.metric_record(router_entry, record) for record in firewall_raw_records_ipv6]     
                firewall_raw_metrics_ipv6 = BaseCollector.counter_collector('firewall_raw_ipv6', 'Total amount of bytes matched by raw firewall rules (IPv6)', metrics_records_ipv6, 'bytes', ['name', 'log'])
                yield firewall_raw_metrics_ipv6

            filter_nat_records_ipv6 = FirewallMetricsDataSource.metric_records(router_entry, metric_labels = firewall_labels, filter_path='nat', ipv6=True)
            if filter_nat_records_ipv6:
                metrics_records_ipv6 = [FirewallCollector.metric_record(router_entry, record) for record in filter_nat_records_ipv6]
                filter_nat_metrics_ipv6 = BaseCollector.counter_collector('firewall_nat_ipv6', 'Total amount of bytes matched by NAT rules (IPv6)', metrics_records_ipv6, 'bytes', ['name', 'log'])
                yield filter_nat_metrics_ipv6

            filter_mangle_records_ipv6 = FirewallMetricsDataSource.metric_records(router_entry, metric_labels = firewall_labels, filter_path='mangle', ipv6=True)
            if filter_mangle_records_ipv6:
                metrics_records_ipv6 = [FirewallCollector.metric_record(router_entry, record) for record in filter_mangle_records_ipv6]
                filter_mangle_metrics_ipv6 = BaseCollector.counter_collector('firewall_mangle_ipv6', 'Total amount of bytes matched by Mangle rules (IPv6)', metrics_records_ipv6, 'bytes', ['name', 'log'])
                yield filter_mangle_metrics_ipv6

    # Helpers
    @staticmethod
    def metric_record(router_entry, firewall_record):
        name = f"| {firewall_record.get('chain', ' ')} | {firewall_record.get('action', ' ')} | {firewall_record.get('comment', ' ')}"
        bytes = firewall_record.get('bytes', 0)
        out_interface = firewall_record.get('out_interface')
        protocol = firewall_record.get('protocol')
        if out_interface:
            name = f"{name} | {out_interface}"
        if protocol:
            name = f"{name} | {protocol}"
        return {MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                'name': name, 'log': firewall_record['log'], 'bytes': bytes}



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


import os
from argparse import ArgumentParser, HelpFormatter
from mktxp.cli.config.config import config_handler, MKTXPConfigKeys, CustomConfig
from mktxp.utils.utils import FSHelper, UniquePartialMatchList, run_cmd
from importlib.metadata import version as Version

class MKTXPCommands:
    INFO = 'info'
    EDIT = 'edit'
    EXPORT = 'export'
    PRINT = 'print'
    SHOW = 'show'

    @classmethod
    def commands_meta(cls):
        return ''.join(('{',
                        f'{cls.INFO}, ',
                        f'{cls.EDIT}, ',
                        f'{cls.EXPORT}, ',
                        f'{cls.PRINT}, ',
                        f'{cls.SHOW}, ',
                        '}'))

class MKTXPOptionsParser:
    ''' Base MKTXP Options Parser
    '''
    def __init__(self):
        self._script_name = f'MKTXP'
        version = Version('mktxp')
        self._description =  \
f'''
Prometheus Exporter for Mikrotik RouterOS, version {version}
Supports gathering metrics across multiple RouterOS devices, all easily configurable via built-in CLI interface.
Comes along with a dedicated Grafana dashboard (https://grafana.com/grafana/dashboards/13679)
Selected metrics info can be printed on the command line. For more information, run: 'mktxp -h'
'''

    @property
    def description(self):
        return self._description

    @property
    def script_name(self):
        return self._script_name

    # Options Parsing Workflow
    def parse_options(self):
        ''' General Options parsing workflow
        '''

        global_options_parser = ArgumentParser(add_help=False)
        self.parse_global_options(global_options_parser)
        namespace, _ = global_options_parser.parse_known_args()    
        if namespace.cfg_dir:
            config_handler(CustomConfig(namespace.cfg_dir))
        else:
            config_handler()

        commands_parser = ArgumentParser(prog = self._script_name,
                                description = 'Prometheus Exporter for Mikrotik RouterOS',
                                formatter_class=MKTXPHelpFormatter, parents=[global_options_parser])
        self.parse_commands(commands_parser)
        args = vars(commands_parser.parse_args())

        self._check_args(args, commands_parser)

        return args

    def parse_global_options(self, parser):
        ''' Parses global options
        '''
        parser.add_argument('--cfg-dir', dest = 'cfg_dir', 
                    type = lambda d: self._is_valid_dir_path(parser, d),
                    help = 'MKTXP config files directory (optional)')

    def parse_commands(self, parser):
        ''' Commands parsing
        '''
        subparsers = parser.add_subparsers(dest = 'sub_cmd',
                                           title = 'MKTXP commands',
                                           metavar = MKTXPCommands.commands_meta())
        
        # Info command
        subparsers.add_parser(MKTXPCommands.INFO,
                                        description = 'Displays MKTXP info',
                                        formatter_class=MKTXPHelpFormatter)
        # Show command
        show_parser = subparsers.add_parser(MKTXPCommands.SHOW,
                                        description = 'Displays MKTXP config router entries',
                                        formatter_class=MKTXPHelpFormatter)
        self._add_entry_name(show_parser, registered_only = True, required = False, help = "Config entry name")
        show_parser.add_argument('-cfg', '--config', dest='config',
                                        help = "Shows MKTXP config files paths",
                                        action = 'store_true')

        # Edit command
        edit_parser = subparsers.add_parser(MKTXPCommands.EDIT,
                                        description = 'Edits an existing MKTXP router entry',
                                        formatter_class=MKTXPHelpFormatter)
        optional_args_group = edit_parser.add_argument_group('Optional Arguments')
        optional_args_group.add_argument('-ed', '--editor', dest='editor',
                help = f"Command line editor to use ({self._system_editor()} by default)",
                default = self._system_editor(),
                type = str)        
        optional_args_group.add_argument('-i', '--internal', dest='internal',
                help = f"Edit MKTXP internal configuration (advanced)",
                action = 'store_true')        

        # Export command
        export_parser = subparsers.add_parser(MKTXPCommands.EXPORT,
                                        description = 'Starts exporting Miktorik Router Metrics to Prometheus',
                                        formatter_class=MKTXPHelpFormatter)

        # Print command
        print_parser = subparsers.add_parser(MKTXPCommands.PRINT,
                                        description = 'Displays selected metrics on the command line',
                                        formatter_class=MKTXPHelpFormatter)
        required_args_group = print_parser.add_argument_group('Required Arguments')
        self._add_entry_name(required_args_group, registered_only = True, help = "Name of config RouterOS entry")

        optional_args_group = print_parser.add_argument_group('Optional Arguments')
        optional_args_group.add_argument('-cc', '--capsman_clients', dest='capsman_clients',
                help = "CAPsMAN clients metrics",
                action = 'store_true')

        optional_args_group.add_argument('-wc', '--wifi_clients', dest='wifi_clients',
                help = "WiFi clients metrics",
                action = 'store_true')

        optional_args_group.add_argument('-dc', '--dhcp_clients', dest='dhcp_clients',
                help = "DHCP clients metrics",
                action = 'store_true')

        optional_args_group.add_argument('-cn', '--conn_stats', dest='conn_stats',
                help = "IP connections stats",
                action = 'store_true')


    # Options checking
    def _check_args(self, args, parser):
        ''' Validation of supplied CLI arguments
        '''
        # check if there is a cmd to execute
        self._check_cmd_args(args, parser)

        if args['sub_cmd'] in (MKTXPCommands.SHOW, MKTXPCommands.PRINT):
            # Registered Entry name could be a partial match, need to expand
            if args['entry_name']:
                args['entry_name'] = UniquePartialMatchList(config_handler.registered_entries()).find(args['entry_name'])

        if args['sub_cmd'] == MKTXPCommands.PRINT:
            if not config_handler.config_entry(args['entry_name']).enabled:
                print(f"Can not print metrics for disabled RouterOS entry: {args['entry_name']}\nRun 'mktxp edit' to review and enable it in the configuration file first")
                parser.exit()

    def _check_cmd_args(self, args, parser):
        ''' Validation of supplied CLI commands
        '''
        # base command check
        if 'sub_cmd' not in args or not args['sub_cmd']:
            # If no command was specified, check for the default one
            cmd = self._default_command
            if cmd:
                args['sub_cmd'] = cmd
            else:
                # no appropriate default either
                parser.print_help()
                parser.exit()


    @property
    def _default_command(self):
        ''' If no command was specified, print INFO by default
        '''
        return MKTXPCommands.INFO


    # Internal helpers
    @staticmethod
    def _is_valid_dir_path(parser, path_arg):
        ''' Checks if path_arg is a valid dir path
        '''
        path_arg = FSHelper.full_path(path_arg)
        if not (os.path.exists(path_arg) and os.path.isdir(path_arg)):
            parser.error(f'"{path_arg}" does not seem to be an existing directory path')
        else:
            return path_arg

    @staticmethod
    def _is_valid_file_path(parser, path_arg):
        ''' Checks if path_arg is a valid file path
        '''
        path_arg = FSHelper.full_path(path_arg)
        if not (os.path.exists(path_arg) and os.path.isfile(path_arg)):
            parser.error('"{path_arg}" does not seem to be an existing file path')
        else:
            return path_arg

    @staticmethod
    def _add_entry_name(parser, registered_only = False, required = True, help = 'MKTXP Entry name'):
        parser.add_argument('-en', '--entry-name', dest = 'entry_name',
            type = str,
            metavar = list(config_handler.registered_entries()) if registered_only else None,
            required = required,
            choices = UniquePartialMatchList(config_handler.registered_entries())if registered_only else None,
            help = help)

    @staticmethod
    def _add_entry_groups(parser):
        required_args_group = parser.add_argument_group('Required Arguments')
        MKTXPOptionsParser._add_entry_name(required_args_group)

    @staticmethod
    def _system_editor():
        editor = os.environ.get('EDITOR')
        if editor:
            return editor

        commands = ['which nano', 'which vi', 'which vim']
        for command in commands:
            editor = run_cmd(command, quiet = True).rstrip()
            if editor:
                break               
        return editor


class MKTXPHelpFormatter(HelpFormatter):
    ''' Custom formatter for ArgumentParser
        Disables double metavar display, showing only for long-named options
    '''
    def _format_action_invocation(self, action):
        if not action.option_strings:
            metavar, = self._metavar_formatter(action, action.dest)(1)
            return metavar
        else:
            parts = []
            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                parts.extend(action.option_strings)

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            # change to
            #    -s, --long ARGS
            else:
                default = action.dest.upper()
                args_string = self._format_args(action, default)
                for option_string in action.option_strings:
                    #parts.append('%s %s' % (option_string, args_string))
                    parts.append('%s' % option_string)
                parts[-1] += ' %s'%args_string
            return ', '.join(parts)




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class QueueMetricsDataSource:
    ''' Queue Metrics data provider
    '''             
    @staticmethod    
    def metric_records(router_entry, *, metric_labels = None, kind = 'tree'):
        if metric_labels is None:
            metric_labels = []                
        try:
            queue_records = router_entry.api_connection.router_api().get_resource(f'/queue/{kind}/').get()
            queue_records = BaseDSProcessor.trimmed_records(router_entry, router_records = queue_records, metric_labels = metric_labels)            
        except Exception as exc:
            print(f'Error getting system resource info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None

        if kind == 'tree':            
            return queue_records

        # simple queue records need splitting upload/download values
        splitted_queue_records = []
        for queue_record in queue_records:
            splitted_queue_record = {}
            for key, value in queue_record.items():
                split_values = value.split('/')
                if split_values and len(split_values) > 1:
                    splitted_queue_record[f'{key}_up'] = split_values[0]
                    splitted_queue_record[f'{key}_down'] = split_values[1]
                else:
                    splitted_queue_record[key] = value
            splitted_queue_records.append(splitted_queue_record)            
        return splitted_queue_records

        





# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.system_resource_ds import SystemResourceMetricsDataSource
from mktxp.utils.utils import check_for_updates


class SystemResourceCollector(BaseCollector):
    ''' System Resource Metrics collector
    '''        
    @staticmethod
    def collect(router_entry):
        resource_labels = ['uptime', 'version', 'free_memory', 'total_memory', 
                           'cpu', 'cpu_count', 'cpu_frequency', 'cpu_load', 
                           'free_hdd_space', 'total_hdd_space', 
                           'architecture_name', 'board_name']
        translation_table = {'uptime': lambda value: BaseOutputProcessor.parse_timedelta_seconds(value)}

        resource_records = SystemResourceMetricsDataSource.metric_records(router_entry, metric_labels = resource_labels, translation_table=translation_table)   
        if resource_records:
            uptime_metrics = BaseCollector.gauge_collector('system_uptime', 'Time interval since boot-up', resource_records, 'uptime', ['version', 'board_name', 'cpu', 'architecture_name'])
            yield uptime_metrics

            free_memory_metrics = BaseCollector.gauge_collector('system_free_memory', 'Unused amount of RAM', resource_records, 'free_memory', ['version', 'board_name', 'cpu', 'architecture_name'])
            yield free_memory_metrics

            total_memory_metrics = BaseCollector.gauge_collector('system_total_memory', 'Amount of installed RAM', resource_records, 'total_memory', ['version', 'board_name', 'cpu', 'architecture_name'])
            yield total_memory_metrics

            free_hdd_metrics = BaseCollector.gauge_collector('system_free_hdd_space', 'Free space on hard drive or NAND', resource_records, 'free_hdd_space', ['version', 'board_name', 'cpu', 'architecture_name'])
            yield free_hdd_metrics

            total_hdd_metrics = BaseCollector.gauge_collector('system_total_hdd_space', 'Size of the hard drive or NAND', resource_records, 'total_hdd_space', ['version', 'board_name', 'cpu', 'architecture_name'])
            yield total_hdd_metrics

            cpu_load_metrics = BaseCollector.gauge_collector('system_cpu_load', 'Percentage of used CPU resources', resource_records, 'cpu_load', ['version', 'board_name', 'cpu', 'architecture_name'])
            yield cpu_load_metrics

            cpu_count_metrics = BaseCollector.gauge_collector('system_cpu_count', 'Number of CPUs present on the system', resource_records, 'cpu_count', ['version', 'board_name', 'cpu', 'architecture_name'])
            yield cpu_count_metrics

            cpu_frequency_metrics = BaseCollector.gauge_collector('system_cpu_frequency', 'Current CPU frequency', resource_records, 'cpu_frequency', ['version', 'board_name', 'cpu', 'architecture_name'])
            yield cpu_frequency_metrics

            # Check for updates
            if router_entry.config_entry.check_for_updates:
                for record in resource_records:
                    cur_version, newest_version = check_for_updates(record['version'])
                    record['newest_version'] = str(newest_version)
                    record['update_available'] = cur_version < newest_version

                update_available_metrics = BaseCollector.gauge_collector('system_update_available', 'Is there a newer version available', resource_records, 'update_available', ['newest_version',])
                yield update_available_metrics



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.queue_ds import QueueMetricsDataSource


class QueueTreeCollector(BaseCollector):
    '''Queue Tree collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.queue:
            return

        qt_labels = ['name', 'parent', 'packet_mark', 'limit_at', 'max_limit', 'priority', 'bytes', 'queued_bytes', 'dropped', 'rate', 'disabled']
        qt_records = QueueMetricsDataSource.metric_records(router_entry, metric_labels=qt_labels, kind = 'tree')

        if qt_records:
            qt_rate_metric = BaseCollector.counter_collector('queue_tree_rates', 'Average passing data rate in bytes per second', qt_records, 'rate', ['name'])
            yield qt_rate_metric

            qt_byte_metric = BaseCollector.counter_collector('queue_tree_bytes', 'Number of processed bytes', qt_records, 'bytes', ['name'])
            yield qt_byte_metric

            qt_queued_metric = BaseCollector.counter_collector('queue_tree_queued_bytes', 'Number of queued bytes', qt_records, 'queued_bytes', ['name'])
            yield qt_queued_metric


            qt_drop_metric = BaseCollector.counter_collector('queue_tree_dropped', 'Number of dropped bytes', qt_records, 'dropped', ['name'])
            yield qt_drop_metric


class QueueSimpleCollector(BaseCollector):
    '''Simple Queue collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.queue:
            return

        qt_labels = ['name', 'parent', 'packet_mark', 'limit_at', 'max_limit', 'priority', 'bytes', 'packets', 'queued_bytes', 'queued_packets','dropped', 'rate', 'packet_rate', 'disabled']
        qt_records = QueueMetricsDataSource.metric_records(router_entry, metric_labels=qt_labels, kind = 'simple')

        if qt_records:
            qt_rate_metric = BaseCollector.counter_collector('queue_simple_rates_upload', 'Average passing upload data rate in bytes per second', qt_records, 'rate_up', ['name'])
            yield qt_rate_metric

            qt_rate_metric = BaseCollector.counter_collector('queue_simple_rates_download', 'Average passing download data rate in bytes per second', qt_records, 'rate_down', ['name'])
            yield qt_rate_metric


            qt_byte_metric = BaseCollector.counter_collector('queue_simple_bytes_upload', 'Number of upload processed bytes', qt_records, 'bytes_up', ['name'])
            yield qt_byte_metric

            qt_byte_metric = BaseCollector.counter_collector('queue_simple_bytes_download', 'Number of download processed bytes', qt_records, 'bytes_down', ['name'])
            yield qt_byte_metric


            qt_queued_metric = BaseCollector.counter_collector('queue_simple_queued_bytes_upload', 'Number of upload queued bytes', qt_records, 'queued_bytes_up', ['name'])
            yield qt_queued_metric


            qt_queued_metric = BaseCollector.counter_collector('queue_simple_queued_bytes_downloadd', 'Number of download queued bytes', qt_records, 'queued_bytes_down', ['name'])
            yield qt_queued_metric


            qt_drop_metric = BaseCollector.counter_collector('queue_simple_dropped_upload', 'Number of upload dropped bytes', qt_records, 'dropped_up', ['name'])
            yield qt_drop_metric


            qt_drop_metric = BaseCollector.counter_collector('queue_simple_dropped_download', 'Number of download dropped bytes', qt_records, 'dropped_down', ['name'])
            yield qt_drop_metric




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor
from mktxp.utils.utils import str2bool

class RouteMetricsDataSource:
    ''' Routes Metrics data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None, ipv6 = False):
        ip_stack = 'ipv6' if ipv6 else 'ip'
        if metric_labels is None:
            metric_labels = []                
        try:
            #route_records = router_entry.api_connection.router_api().get_resource(f'/{ip_stack}/route').get(active='yes')
            route_records = router_entry.api_connection.router_api().get_resource(f'/{ip_stack}/route').call('print', {'proplist':'active,connect,dynamic,static,bgp,ospf'})            
            
            #active_records = [record for record in route_records if record.get('active')]
            RouteMetricsDataSource._remove_from_list_of_dict(route_records, 'active')

            return BaseDSProcessor.trimmed_records(router_entry, router_records = route_records, metric_labels = metric_labels)
        except Exception as exc:
            print(f'Error getting {"IPv6" if ipv6 else "IPv4"} routes info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None

    # helpers
    @staticmethod
    def _remove_from_list_of_dict(dict_list, key):
        indexes = []
        for index, dict in enumerate(dict_list):
            if not str2bool(dict.get(key)):
                indexes.append(index)
        offset = 0
        for index in indexes:
            dict_list.pop(index-offset)
            offset += 1        



#!/usr/bin/env python
# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

''' Python version check
'''

from __future__ import print_function

import sys

def check_version():
    if sys.version_info.major < 3:
        print(\
        '''
        Mikrotik Prometheus Exporter requires
                           Python version 3.8 or later.

        You can create an isolated Python 3.8 environment
        with the virtualenv tool:
          http://docs.python-guide.org/en/latest/dev/virtualenvs

        ''')
        sys.exit(0)
    elif sys.version_info.major == 3 and sys.version_info.minor < 8:
        print(\
        '''

        Mikrotik Prometheus Exporter requires
                            Python version 3.8 or later.

        Please upgrade to the latest Python 3.x version.

        ''')
        sys.exit(0)

# check
check_version()



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


import re, os
from datetime import timedelta
from collections import namedtuple
from texttable import Texttable
from humanize import naturaldelta
from mktxp.cli.config.config import config_handler
from mktxp.datasource.wireless_ds import WirelessMetricsDataSource
from mktxp.datasource.dhcp_ds import DHCPMetricsDataSource
from math import floor, log


class BaseOutputProcessor:
    OutputCapsmanEntry = namedtuple('OutputCapsmanEntry', ['dhcp_name', 'dhcp_address', 'mac_address', 'rx_signal', 'interface', 'ssid', 'tx_rate', 'rx_rate', 'uptime'])
    OutputCapsmanEntry.__new__.__defaults__ = ('',) * len(OutputCapsmanEntry._fields)

    OutputWirelessEntry = namedtuple('OutputWirelessEntry', ['dhcp_name', 'dhcp_address', 'mac_address', 'signal_strength', 'signal_to_noise', 'interface', 'tx_rate', 'rx_rate', 'uptime'])
    OutputWirelessEntry.__new__.__defaults__ = ('',) * len(OutputWirelessEntry._fields)

    OutputWiFiEntry = namedtuple('OutputWiFiEntry', ['dhcp_name', 'dhcp_address', 'mac_address', 'signal_strength', 'interface', 'tx_rate', 'rx_rate', 'uptime'])
    OutputWiFiEntry.__new__.__defaults__ = ('',) * len(OutputWiFiEntry._fields)

    OutputDHCPEntry = namedtuple('OutputDHCPEntry', ['host_name', 'server', 'mac_address', 'address', 'active_address', 'expires_after'])
    OutputDHCPEntry.__new__.__defaults__ = ('',) * len(OutputDHCPEntry._fields)

    OutputConnStatsEntry = namedtuple('OutputConnStatsEntry', ['dhcp_name', 'src_address', 'connection_count', 'dst_addresses'])
    OutputConnStatsEntry.__new__.__defaults__ = ('',) * len(OutputConnStatsEntry._fields)


    @staticmethod
    def augment_record(router_entry, registration_record, id_key = 'mac_address'):
        BaseOutputProcessor.resolve_dhcp(router_entry, registration_record, id_key)

        # split out tx/rx bytes
        if registration_record.get('bytes'):
            registration_record['tx_bytes'] = registration_record['bytes'].split(',')[0]
            registration_record['rx_bytes'] = registration_record['bytes'].split(',')[1]
            del registration_record['bytes']

        if registration_record.get('tx_rate'):
            registration_record['tx_rate'] = BaseOutputProcessor.parse_bitrates(registration_record['tx_rate'])
        if registration_record.get('rx_rate'):
            registration_record['rx_rate'] = BaseOutputProcessor.parse_bitrates(registration_record['rx_rate'])
        if registration_record.get('uptime'):
            registration_record['uptime'] = naturaldelta(BaseOutputProcessor.parse_timedelta_seconds(registration_record['uptime']), months=True, minimum_unit='seconds')

        if registration_record.get('signal_strength'):
            registration_record['signal_strength'] = BaseOutputProcessor.parse_signal_strength(registration_record['signal_strength'])
        if registration_record.get('rx_signal'):
            registration_record['rx_signal'] = BaseOutputProcessor.parse_signal_strength(registration_record['rx_signal'])

    @staticmethod
    def dhcp_name(router_entry, dhcp_lease_record, drop_comment = False):
        dhcp_name = dhcp_lease_record.get('host_name')
        dhcp_comment = dhcp_lease_record.get('comment')
        
        if dhcp_name and dhcp_comment:
            dhcp_name = f'{dhcp_name[0:20]} ({dhcp_comment[0:20]})' if not router_entry.config_entry.use_comments_over_names else dhcp_comment
        elif dhcp_comment:
            dhcp_name = dhcp_comment
        else:
            dhcp_name = dhcp_lease_record.get('mac_address') if not dhcp_name else dhcp_name        

        if drop_comment:
            del dhcp_lease_record['comment']

        return dhcp_name

    @staticmethod
    def resolve_dhcp(router_entry, registration_record, id_key = 'mac_address', resolve_address = True):
        if not router_entry.dhcp_records:
            DHCPMetricsDataSource.metric_records(router_entry)
        dhcp_name = registration_record.get(id_key)
        dhcp_address = 'No DHCP Record'              

        dhcp_lease_record = router_entry.dhcp_record(dhcp_name)
        if dhcp_lease_record:
            dhcp_name = BaseOutputProcessor.dhcp_name(router_entry, dhcp_lease_record)
            dhcp_address = dhcp_lease_record.get('address', '')

        registration_record['dhcp_name'] = dhcp_name
        if resolve_address:
            registration_record['dhcp_address'] = dhcp_address

    @staticmethod
    def parse_rates(rate):
        rates_rgx = config_handler.re_compiled.get('rates_rgx')
        if not rates_rgx:
            rates_rgx = re.compile(r'(\d*(?:\.\d*)?)([GgMmKk]bps?)')
            config_handler.re_compiled['rates_rgx'] = rates_rgx
        rc = rates_rgx.search(rate)
        return f'{int(float(rc[1]))} {rc[2]}' if rc and len(rc.groups()) == 2 else rate

    @staticmethod
    def parse_bitrates(rate):
        try:
            rate = int(rate)
        except:
            return BaseOutputProcessor.parse_rates(rate)
        power = floor(log(rate, 1000))
        return f"{int(rate / 1000 ** power)} {['bps', 'Kbps', 'Mbps', 'Gbps'][int(power)]}"

    @staticmethod
    def parse_timedelta(time, ms_span=False):
        # ms_span for milliseconds-long durations, since otherwise minutes would match the ms in the value
        rgx_key = 'duration_interval_rgx_sp' if ms_span else 'duration_interval_rgx'
        duration_interval_rgx = config_handler.re_compiled.get(rgx_key)
        if not duration_interval_rgx:
            duration_interval_rgx = re.compile(r'((?P<seconds>\d+)s)?((?P<milliseconds>\d+)ms)?((?P<microseconds>\d+)us)?') if ms_span else\
                    re.compile(r'((?P<weeks>\d+)w)?((?P<days>\d+)d)?((?P<hours>\d+)h)?((?P<minutes>\d+)m)?((?P<seconds>\d+)s)?((?P<milliseconds>\d+)ms)?')
            config_handler.re_compiled[rgx_key] = duration_interval_rgx                        
        time_dict = duration_interval_rgx.match(time).groupdict()
        return timedelta(**{key: int(value) for key, value in time_dict.items() if value})

    @staticmethod
    def parse_timedelta_seconds(time, ms_span=False):
        return BaseOutputProcessor.parse_timedelta(time, ms_span=ms_span).total_seconds()

    @staticmethod
    def parse_timedelta_milliseconds(time, ms_span=False):
        return BaseOutputProcessor.parse_timedelta(time, ms_span=ms_span) / timedelta(milliseconds=1)

    @staticmethod
    def parse_signal_strength(signal_strength):
        wifi_signal_strength_rgx = config_handler.re_compiled.get('wifi_signal_strength_rgx')
        if not wifi_signal_strength_rgx:
            # wifi_signal_strength_rgx = re.compile(r'(-?\d+(?:\.\d+)?)(dBm)?')
            wifi_signal_strength_rgx = re.compile(r'(-?\d+(?:\.\d+)?)')           
            config_handler.re_compiled['wifi_signal_strength_rgx'] = wifi_signal_strength_rgx
        return wifi_signal_strength_rgx.search(signal_strength).group()

    @staticmethod
    def parse_interface_rate(interface_rate):
        interface_rate_rgx = config_handler.re_compiled.get('interface_rate_rgx')
        if not interface_rate_rgx:
            interface_rate_rgx = re.compile(r'[^.\-\d]')
            config_handler.re_compiled['interface_rate_rgx'] = interface_rate_rgx
        rate = lambda interface_rate: 1000 if interface_rate.find('Mbps') < 0 else 1
        return(int(float(interface_rate_rgx.sub('', interface_rate)) * rate(interface_rate)))

    @staticmethod
    def output_table(outputEntry = None):
        table = Texttable(max_width = os.get_terminal_size().columns)
        table.set_deco(Texttable.HEADER | Texttable.BORDER | Texttable.VLINES )        
        if outputEntry:
            table.header(outputEntry._fields)
            table.set_cols_align(['l']+ ['c']*(len(outputEntry._fields)-1))
        return table



# coding=utf8
# Copyright (c) 2020 Arseniy Kuznetsov
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor
from mktxp.flow.router_entry import RouterEntry

TRANSLATION_TABLE = {
    'comment': lambda value: value if value else '',
    'log': lambda value: '1' if value == 'true' else '0'
}

class FirewallMetricsDataSource:
    ''' Firewall Metrics data provider, supports both IPv4 and IPv6
    '''
    @staticmethod
    def metric_records(router_entry, *, metric_labels=None, matching_only=True, filter_path='filter', ipv6 = False):
        if metric_labels is None:
            metric_labels = []
        try:
            ip_stack = 'ipv6' if ipv6 else 'ip'
            filter_paths = {
                'filter': f'/{ip_stack}/firewall/filter',
                'raw': f'/{ip_stack}/firewall/raw',
                'nat': f'/{ip_stack}/firewall/nat',
                'mangle': f'/{ip_stack}/firewall/mangle'
            }
            filter_path = filter_paths[filter_path]
            firewall_records = FirewallMetricsDataSource._get_records(
                router_entry,
                filter_path,
                {'stats': ''} if ipv6 else {'stats': '', 'all': ''},
                matching_only=matching_only
            )

            return BaseDSProcessor.trimmed_records(router_entry, router_records=firewall_records, metric_labels=metric_labels, translation_table=TRANSLATION_TABLE)
        except Exception as exc:
            print(
                f'Error getting {"IPv6" if ipv6 else "IPv4"} firewall filters info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}'
            )
            return None

    # helpers
    @staticmethod
    def _get_records(router_entry: RouterEntry, filter_path: str, args: dict, matching_only: bool = False):
        """
        Get firewall records from a Mikrotik ROS device.
        :param router_entry: The ROS API entry used to connect to the API
        :param filter_path:  The path to query the records for (e.g. /ip/firewall/filter)
        :param args:         A dictionary of arguments to pass to the print function used for export.
                             Looks like: '{'stats': '', 'all': ''}'
        """
        firewall_records = router_entry.api_connection.router_api().get_resource(filter_path).call('print', args)
        if matching_only:
            firewall_records = [record for record in firewall_records if int(record.get('bytes', '0')) > 0]
        return firewall_records



# coding=utf8
# Copyright (c) 2020 Arseniy Kuznetsov
##
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
##
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.container_ds import ContainerDataSource
from mktxp.datasource.package_ds import PackageMetricsDataSource


class ContainerCollector(BaseCollector):
    '''Neighbor Collector'''

    @staticmethod
    def collect(router_entry):
        if router_entry.config_entry.container and PackageMetricsDataSource.is_package_installed(router_entry, package_name='container'):
            metric_labels = ['name', 'repo', 'os', 'arch', 'status']
            records = ContainerDataSource.metric_records(router_entry, metric_labels=metric_labels)
            metrics = BaseCollector.info_collector('container', 'Containers', records, metric_labels=metric_labels)
            yield metrics



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.package_ds import PackageMetricsDataSource


class PackageCollector(BaseCollector):
    '''Installed Packages collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.installed_packages:
            return

        package_labels = ['name', 'version', 'build_time', 'disabled']
        package_records = PackageMetricsDataSource.metric_records(router_entry, metric_labels=package_labels)
        if package_records:
            package_metrics = BaseCollector.info_collector('installed_packages', 'Installed Packages', package_records, package_labels)
            yield package_metrics




# coding=utf8
# Copyright (c) 2020 Arseniy Kuznetsov
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class DNSDataSource:
    def metric_records(router_entry):
        try:
            router_records = router_entry.api_connection.router_api().get_resource('/ip/dns').get()
            dns_records = BaseDSProcessor.trimmed_records(router_entry, router_records=router_records)
            return dns_records[0] if dns_records else None
        except Exception as exc:
            print(f'Error getting DNS info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor
from mktxp.datasource.wireless_ds import WirelessMetricsDataSource
from mktxp.flow.router_entry import RouterEntryWirelessType

class CapsmanInfo:
    @staticmethod
    def capsman_paths(router_entry):
        if router_entry.capsman_entry.wireless_type == RouterEntryWirelessType.DUAL:
            return ['/caps-man', f'/interface/wifi/capsman']
        elif router_entry.capsman_entry.wireless_type == RouterEntryWirelessType.WIRELESS:
            return ['/caps-man']
        else:    
            wireless_package = WirelessMetricsDataSource.wireless_package(router_entry.capsman_entry)
            return [f'/interface/{wireless_package}/capsman']

    @staticmethod
    def registration_table_paths(router_entry):
        if router_entry.capsman_entry.wireless_type == RouterEntryWirelessType.DUAL:
            return ['/caps-man/registration-table', f'/interface/wifi/registration-table']
        elif router_entry.capsman_entry.wireless_type == RouterEntryWirelessType.WIRELESS:
            return ['/caps-man/registration-table']
        else:    
            wireless_package = WirelessMetricsDataSource.wireless_package(router_entry.capsman_entry)
            return [f'/interface/{wireless_package}/registration-table']


class CapsmanCapsMetricsDataSource:
    ''' Caps Metrics data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None):
        if metric_labels is None:
            metric_labels = []                
        try:
            remote_caps_records = []
            for capsman_path in CapsmanInfo.capsman_paths(router_entry.capsman_entry):
                remote_caps_records.extend(router_entry.capsman_entry.api_connection.router_api().get_resource(f'{capsman_path}/remote-cap').get())
            return BaseDSProcessor.trimmed_records(router_entry, router_records = remote_caps_records, metric_labels = metric_labels)
        except Exception as exc:
            print(f'Error getting CAPsMAN remote caps info from router {router_entry.capsman_entry.router_name}@{router_entry.capsman_entry.config_entry.hostname}: {exc}')
            return None

class CapsmanRegistrationsMetricsDataSource:
    ''' Capsman Registrations Metrics data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None,  add_router_id = True):
        if metric_labels is None:
            metric_labels = []                
        try:
            registration_table_records = []
            for registration_table_path in CapsmanInfo.registration_table_paths(router_entry.capsman_entry):
                registration_table_records.extend(router_entry.capsman_entry.api_connection.router_api().get_resource(f'{registration_table_path}').get())
            
            # With wifiwave2, Mikrotik renamed the field 'rx-signal' to 'signal' 
            # For backward compatibility, including both variants
            for record in registration_table_records:
                if 'signal' in record:
                    record['rx-signal'] = record['signal']

            return BaseDSProcessor.trimmed_records(router_entry, router_records = registration_table_records, metric_labels = metric_labels, add_router_id = add_router_id)
        except Exception as exc:
            print(f'Error getting CAPsMAN registration table info from router {router_entry.capsman_entry.router_name}@{router_entry.capsman_entry.config_entry.hostname}: {exc}')
            return None


class CapsmanInterfacesDatasource:
    ''' Data provider for CAPsMaN interfaces
    '''
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None):
        if not router_entry.capsman_entry.wireless_type in (RouterEntryWirelessType.DUAL, RouterEntryWirelessType.WIRELESS):
            return None            
        if metric_labels is None:
            metric_labels = []                
        try:
            caps_interfaces = router_entry.capsman_entry.api_connection.router_api().get_resource('/caps-man/interface').get()
            return BaseDSProcessor.trimmed_records(router_entry, router_records = caps_interfaces, metric_labels = metric_labels)
        except Exception as exc:
            print(f'Error getting CAPsMAN interfaces info from router {router_entry.capsman_entry.router_name}@{router_entry.capsman_entry.config_entry.hostname}: {exc}')
            return None



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor
from mktxp.datasource.system_resource_ds import SystemResourceMetricsDataSource
from mktxp.utils.utils import routerOS7_version


class BaseInterfaceDataSource:
    @staticmethod
    def rewrite_interface_names(router_entry, metric_records):
        for metric_record in metric_records:
            if metric_record.get('comment'):
                if router_entry.config_entry.use_comments_over_names:
                    metric_record['name'] = metric_record['comment']
                else:
                    metric_record['name'] = f"{metric_record['name']} ({metric_record['comment']})"

        return metric_records


class InterfaceMetricsDataSource(BaseInterfaceDataSource):
    """ Interface Monitor Metrics data provider
    """
    @staticmethod
    def metric_records(router_entry, *, metric_labels=None, kind='ethernet', additional_proplist=None,
                       translation_table=None):
        if metric_labels is None:
            metric_labels = []

        if additional_proplist is None:
            additional_proplist = []

        call_params = {
            'proplist': ','.join(['name', 'running', 'disabled'] + additional_proplist)
        }

        try:
            interface_records = router_entry.api_connection.router_api().get_resource(
                f'/interface/{kind}'
            ).call(
                'print',
                call_params
            )
            interface_records = BaseInterfaceDataSource.rewrite_interface_names(router_entry, interface_records)
            return BaseDSProcessor.trimmed_records(
                router_entry,
                router_records=interface_records,
                metric_labels=metric_labels,
                translation_table=translation_table
            )
        except Exception as exc:
            print(f'Error getting {kind} interface info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None


class InterfaceTrafficMetricsDataSource(BaseInterfaceDataSource):
    """ Interface Traffic Metrics data provider
    """
    @staticmethod
    def metric_records(router_entry, *, metric_labels, translation_table=None):
        metric_labels = metric_labels or []
        try:
            # get stats for all existing interfaces
            metric_stats_records = router_entry.api_connection.router_api().get_resource(
                '/interface'
            ).call(
                'print',
                {'stats': 'detail'}
            )
            metric_stats_records = BaseInterfaceDataSource.rewrite_interface_names(router_entry, metric_stats_records)
            return BaseDSProcessor.trimmed_records(
                router_entry=router_entry,
                router_records=metric_stats_records,
                metric_labels=metric_labels,
                translation_table=translation_table,
            )
        except Exception as exc:
            print(f'Error getting interface traffic stats info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None


class InterfaceMonitorMetricsDataSource:
    """ Interface Monitor Metrics data provider
    """
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None, translation_table = None, kind = 'ethernet', include_comments = False, running_only = True):
        if metric_labels is None:
            metric_labels = []

        ver = SystemResourceMetricsDataSource.os_version(router_entry)
        # On RouterOS 7, the 'monitor' action was replaced with 'info' for LTE interfaces, and the 'number' key was replaced with 'numbers'
        monitor_action = 'info' if kind == 'lte' and not routerOS7_version(ver) else 'monitor'
        monitor_numbers_key = 'number' if kind == 'lte' and not routerOS7_version(ver) else 'numbers'

        try:
            interfaces = router_entry.api_connection.router_api().get_resource(f'/interface/{kind}').call('print', {'proplist':'name,comment,running'})

            interface_monitor_records = []
            for int_num, interface in enumerate(interfaces):
                interface_monitor_record = {}
                if not running_only or interface['running'] == 'true':
                    interface_monitor_record = router_entry.api_connection.router_api().get_resource(f'/interface/{kind}').call(monitor_action, {'once':'', monitor_numbers_key:f'{int_num}'})[0]
                else:
                    # unless explicitly requested, no need to do a monitor call for not running interfaces
                    interface_monitor_record = {'name': interface['name'], 'status': 'no-link'}

                if include_comments and interface.get('comment'):
                    # combines names with comments
                    interface_monitor_record['name'] = interface['comment'] if router_entry.config_entry.use_comments_over_names else \
                                                                                                        f"{interface['name']} ({interface['comment']})"
                interface_monitor_records.append(interface_monitor_record)

            # With wifiwave2, Mikrotik renamed the field 'registered-clients' to 'registered-peers'
            # For backward compatibility, including both variants
            for interface_monitor_record in interface_monitor_records:
                if 'registered-peers' in interface_monitor_record:
                    interface_monitor_record['registered-clients'] = interface_monitor_record['registered-peers']
            return BaseDSProcessor.trimmed_records(router_entry, router_records = interface_monitor_records, metric_labels = metric_labels, translation_table=translation_table)
        except Exception as exc:
            print(f'Error getting {kind} interface monitor info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.wireless_ds import WirelessMetricsDataSource
from mktxp.datasource.interface_ds import InterfaceMonitorMetricsDataSource


class WLANCollector(BaseCollector):
    ''' Wireless Metrics collector
    '''    
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.wireless:
            return

        monitor_labels = ['channel', 'noise_floor', 'overall_tx_ccq', 'registered_clients', 'registered_peers']
        monitor_records = InterfaceMonitorMetricsDataSource.metric_records(router_entry, metric_labels = monitor_labels, kind = WirelessMetricsDataSource.wireless_package(router_entry))   
        if monitor_records:
            # sanitize records for relevant labels
            noise_floor_records = [monitor_record for monitor_record in monitor_records if monitor_record.get('noise_floor')]
            tx_ccq_records = [monitor_record for monitor_record in monitor_records if monitor_record.get('overall_tx_ccq')]
            registered_clients_records  = [monitor_record for monitor_record in monitor_records if monitor_record.get('registered_clients')]

            if noise_floor_records:
                noise_floor_metrics = BaseCollector.gauge_collector('wlan_noise_floor', 'Noise floor threshold', noise_floor_records, 'noise_floor', ['channel'])
                yield noise_floor_metrics

            if tx_ccq_records:
                overall_tx_ccq_metrics = BaseCollector.gauge_collector('wlan_overall_tx_ccq', 'Client Connection Quality for transmitting', tx_ccq_records, 'overall_tx_ccq', ['channel'])
                yield overall_tx_ccq_metrics

            if registered_clients_records:
                registered_clients_metrics = BaseCollector.gauge_collector('wlan_registered_clients', 'Number of registered clients', registered_clients_records, 'registered_clients', ['channel'])
                yield registered_clients_metrics

        # the client info metrics
        if router_entry.config_entry.wireless_clients:
            registration_labels = ['interface', 'ssid', 'mac_address', 'tx_rate', 'rx_rate', 'uptime', 'bytes', 'signal_to_noise', 'tx_ccq', 'signal_strength', 'signal']
            registration_records = WirelessMetricsDataSource.metric_records(router_entry, metric_labels = registration_labels)
            if registration_records:
                for registration_record in registration_records:
                    BaseOutputProcessor.augment_record(router_entry, registration_record)

                tx_byte_metrics = BaseCollector.counter_collector('wlan_clients_tx_bytes', 'Number of sent packet bytes', registration_records, 'tx_bytes', ['dhcp_name', 'mac_address'])
                yield tx_byte_metrics

                rx_byte_metrics = BaseCollector.counter_collector('wlan_clients_rx_bytes', 'Number of received packet bytes', registration_records, 'rx_bytes', ['dhcp_name', 'mac_address'])
                yield rx_byte_metrics

                signal_strength_metrics = BaseCollector.gauge_collector('wlan_clients_signal_strength', 'Average strength of the client signal recevied by AP', registration_records, 'signal_strength', ['dhcp_name', 'mac_address'])
                yield signal_strength_metrics

                signal_to_noise_metrics = BaseCollector.gauge_collector('wlan_clients_signal_to_noise', 'Client devices signal to noise ratio', registration_records, 'signal_to_noise', ['dhcp_name', 'mac_address'])
                yield signal_to_noise_metrics

                tx_ccq_metrics = BaseCollector.gauge_collector('wlan_clients_tx_ccq', 'Client Connection Quality (CCQ) for transmit', registration_records, 'tx_ccq', ['dhcp_name', 'mac_address'])
                yield tx_ccq_metrics

                registration_metrics = BaseCollector.info_collector('wlan_clients_devices', 'Client devices info', 
                                        registration_records, ['dhcp_name', 'dhcp_address', 'rx_signal', 'ssid', 'tx_rate', 'rx_rate', 'interface', 'mac_address', 'uptime'])
                yield registration_metrics





# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.poe_ds import POEMetricsDataSource


class POECollector(BaseCollector):
    ''' POE Metrics collector
    '''    
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.poe:
            return

        poe_labels = ['name', 'poe_out', 'poe_priority', 'poe_voltage', 'poe_out_status', 'poe_out_voltage', 'poe_out_current', 'poe_out_power']
        poe_records = POEMetricsDataSource.metric_records(router_entry, include_comments = True, metric_labels = poe_labels)  

        if poe_records:
            poe_info_labels = ['name', 'poe_out', 'poe_priority', 'poe_voltage', 'poe_out_status']
            poe_metrics = BaseCollector.info_collector('poe', 'POE Info Metrics', poe_records, poe_info_labels)
            yield poe_metrics

            for poe_record in poe_records:
                if 'poe_out_voltage' in poe_record:
                    poe_voltage_metrics = BaseCollector.gauge_collector('poe_out_voltage', 'POE Out Voltage', [poe_record, ], 'poe_out_voltage', ['name'])
                    yield poe_voltage_metrics

                if 'poe_out_current' in poe_record:
                    poe_current_metrics = BaseCollector.gauge_collector('poe_out_current', 'POE Out Current', [poe_record, ], 'poe_out_current', ['name'])
                    yield poe_current_metrics

                if 'poe_out_power' in poe_record:
                    poe_power_metrics = BaseCollector.gauge_collector('poe_out_power', 'POE Out Power', [poe_record, ], 'poe_out_power', ['name'])
                    yield poe_power_metrics




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from mktxp.datasource.base_ds import BaseDSProcessor

class PublicIPAddressDatasource:
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None):
        if metric_labels is None:
            metric_labels = []

        try:
            records = router_entry.api_connection.router_api().get_resource('/ip/cloud/').get()
            return BaseDSProcessor.trimmed_records(router_entry, router_records=records, metric_labels = metric_labels)
        except Exception as exc:
            print(f'Error public IP address info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None


# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.dhcp_ds import DHCPMetricsDataSource


class DHCPOutput:
    ''' DHCP Clients CLI Output
    '''    
    @staticmethod
    def clients_summary(router_entry):
        dhcp_lease_labels = ['host_name', 'comment', 'active_address', 'address', 'mac_address', 'server', 'expires_after']
        dhcp_lease_records = DHCPMetricsDataSource.metric_records(router_entry, metric_labels = dhcp_lease_labels, add_router_id = False, translate = False, dhcp_cache = False)
        if not dhcp_lease_records:
            print('No DHCP registration records')
            return 

        dhcp_by_server = {}
        for dhcp_lease_record in sorted(dhcp_lease_records, key = lambda dhcp_record: dhcp_record['address'], reverse=True):
            server = dhcp_lease_record.get('server', 'all')
            if server == 'all':
                dhcp_lease_record['server'] = server
            if server in dhcp_by_server.keys():
                dhcp_by_server[server].append(dhcp_lease_record)
            else:
                dhcp_by_server[server] = [dhcp_lease_record]         

        output_records = 0
        lease_records = len(dhcp_lease_records)        
        output_entry = BaseOutputProcessor.OutputDHCPEntry
        output_table = BaseOutputProcessor.output_table(output_entry)
                
        for key in dhcp_by_server.keys():
            for record in dhcp_by_server[key]:
                record['host_name'] = BaseOutputProcessor.dhcp_name(router_entry, record, drop_comment = True)
                output_table.add_row(output_entry(**record))
                output_records += 1
            if output_records < lease_records:
                output_table.add_row(output_entry())

        print (output_table.draw())

        for server in dhcp_by_server.keys():
            print(f'{server} clients: {len(dhcp_by_server[server])}')
        print(f'Total DHCP clients: {output_records}', '\n')



# coding=utf8
## Copyright (c) 2024 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.interface_ds import InterfaceMetricsDataSource
from mktxp.utils.utils import parse_mkt_uptime


class EOIPCollector(BaseCollector):
    """ EoIP Metrics collector
    """
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.eoip:
            return

        default_labels = ['name', 'local_address', 'remote_address', 'tunnel_id']
        monitor_records = InterfaceMetricsDataSource.metric_records(
            router_entry,
            kind='eoip',
            additional_proplist=['actual-mtu', 'l2mtu', 'local-address', 'mtu', 'remote-address', 'tunnel-id'],
        )

        if monitor_records:
            yield BaseCollector.gauge_collector(
                'interface_l2mtu',
                'Current used layer 2 mtu for this interface',
                monitor_records,
                metric_key='actual_mtu',
                metric_labels=default_labels
            )

            yield BaseCollector.gauge_collector(
                'interface_mtu',
                'Current used mut for this interface',
                monitor_records,
                metric_key='actual_mtu',
                metric_labels=default_labels + ['mtu']
            )



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class PackageMetricsDataSource:
    ''' Wireless Metrics data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None, add_router_id = True):
        if metric_labels is None:
            metric_labels = []                
        try:
            package_records = router_entry.api_connection.router_api().get_resource('/system/package').get()
            return BaseDSProcessor.trimmed_records(router_entry, router_records = package_records, metric_labels = metric_labels, add_router_id = add_router_id)
        except Exception as exc:
            print(f'Error getting installed packages info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None


    @staticmethod
    def is_package_installed (router_entry, package_name = None, enabled_only = True):
        if package_name:
            try:
                get_params = {'disabled': 'false'} if enabled_only else {}        
                package_records = router_entry.api_connection.router_api().get_resource('/system/package').get(**get_params)
                return any(pkg['name'] == package_name for pkg in package_records)                                             
            except Exception as exc:
                print(f'Error getting an installed package status from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')        
        return False



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.mktxp_ds import MKTXPMetricsDataSource


class MKTXPCollector(BaseCollector):
    ''' System Identity Metrics collector
    '''     
    @staticmethod
    def collect(router_entry):
        mktxp_records = MKTXPMetricsDataSource.metric_records(router_entry)
        if mktxp_records:
            mktxp_duration_metric = BaseCollector.counter_collector('collection_time', 'Total time spent collecting metrics in milliseconds', mktxp_records, 'duration', ['name'])
            yield mktxp_duration_metric





# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.interface_ds import InterfaceMonitorMetricsDataSource


class MonitorCollector(BaseCollector):
    """ Ethernet Interface Monitor Metrics collector
    """
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.monitor:
            return

        monitor_labels = [
            'full_duplex',
            'name',
            'rate',
            'sfp_connector_type',
            'sfp_manufacturing_date',
            'sfp_module_present',
            'sfp_rx_loss',
            'sfp_rx_power',
            'sfp_supply_voltage',
            'sfp_temperature',
            'sfp_tx_bias_current',
            'sfp_tx_fault',
            'sfp_tx_power',
            'sfp_type',
            'sfp_vendor_name',
            'sfp_vendor_part_number',
            'sfp_vendor_revision',
            'sfp_vendor_serial',
            'sfp_wavelength',
            'status',
        ]

        translation_table = {
            'full_duplex': lambda value: '1' if value == 'true' else '0',
            'name': lambda value: value if value else '',
            'rate': lambda value: MonitorCollector._rates(value) if value else '0',
            'sfp_module_present': lambda value: '1' if value == 'true' else '0',
            'sfp_rx_loss': lambda value: '1' if value == 'true' else '0',
            'sfp_temperature': lambda value: value if value else '0',
            'sfp_tx_bias_current': lambda value: float(value) / 1000 if value else '0',
            'sfp_tx_fault': lambda value: '1' if value == 'true' else '0',
            'status': lambda value: '1' if value == 'link-ok' else '0',
        }
        monitor_records = InterfaceMonitorMetricsDataSource.metric_records(
            router_entry,
            metric_labels=monitor_labels,
            translation_table=translation_table,
            include_comments=True
        )

        if not monitor_records:
            return

        yield BaseCollector.gauge_collector(
            'interface_status',
            'Current interface link status',
            monitor_records,
            'status',
            ['name']
        )

        # limit records according to the relevant metrics
        rate_records = [monitor_record for monitor_record in monitor_records if monitor_record.get('rate', None)]
        yield BaseCollector.gauge_collector(
            'interface_rate',
            'Actual interface connection data rate',
            rate_records,
            'rate',
            ['name']
        )

        full_duplex_records = [monitor_record for monitor_record in monitor_records if monitor_record.get('full_duplex', None)]
        yield BaseCollector.gauge_collector(
            'interface_full_duplex',
            'Full duplex data transmission',
            full_duplex_records,
            'full_duplex',
            ['name']
        )

        sfp_metrics = [record for record in monitor_records if int(record.get("sfp_module_present"))]
        if sfp_metrics:
            yield BaseCollector.info_collector(
                'interface_sfp',
                'Information about the used transceiver',
                sfp_metrics,
                metric_labels=[
                    'name',
                    'sfp_connector_type',
                    'sfp_manufacturing_date',
                    'sfp_type',
                    'sfp_vendor_name',
                    'sfp_vendor_part_number',
                    'sfp_vendor_revision',
                    'sfp_vendor_serial',
                ]
            )
            yield BaseCollector.gauge_collector(
                'interface_sfp_supply_voltage',
                'The transceivers current supply voltage',
                sfp_metrics,
                metric_key='sfp_supply_voltage',
                metric_labels=['name']
            )
            yield BaseCollector.gauge_collector(
                'interface_sfp_rx_loss',
                'The receiver signal is lost',
                sfp_metrics,
                metric_key='sfp_rx_loss',
                metric_labels=['name']
            )
            yield BaseCollector.gauge_collector(
                'interface_sfp_rx_power',
                'Current SFP RX Power',
                sfp_metrics,
                metric_key='sfp_rx_power',
                metric_labels=['name']
            )
            yield BaseCollector.gauge_collector(
                'interface_sfp_temperature',
                'Current SFP Temperature',
                sfp_metrics,
                metric_key='sfp_temperature',
                metric_labels=['name']
            )
            yield BaseCollector.gauge_collector(
                'interface_sfp_tx_bias_current',
                'The transceivers current tx bias current',
                sfp_metrics,
                metric_key='sfp_tx_bias_current',
                metric_labels=['name']
            )
            yield BaseCollector.gauge_collector(
                'interface_sfp_tx_power',
                'Current SFP TX Power',
                sfp_metrics,
                metric_key='sfp_tx_power',
                metric_labels=['name']
            )
            yield BaseCollector.gauge_collector(
                'interface_sfp_tx_fault',
                'The transceiver transmitter is in fault state',
                sfp_metrics,
                metric_key='sfp_tx_fault',
                metric_labels=['name']
            )
            yield BaseCollector.gauge_collector(
                'interface_sfp_wavelength',
                'Current SFP Wavelength',
                sfp_metrics,
                metric_key='sfp_wavelength',
                metric_labels=['name']
            )

    @staticmethod
    def _rates(rate_option):
        # according mikrotik docs, an interface rate should be one of these
        rate_value = {
                '10Mbps': '10',
                '100Mbps': '100',
                '1Gbps': '1000',
                '2.5Gbps': '2500',
                '5Gbps': '5000',
                '10Gbps': '10000',
                '40Gbps': '40000'
                }.get(rate_option, None)
        if rate_value:
            return rate_value

        # ...or just calculate in case it's not
        return BaseOutputProcessor.parse_interface_rate(rate_option)



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.connection_ds import IPConnectionDatasource, IPConnectionStatsDatasource


class IPConnectionCollector(BaseCollector):
    ''' IP Connection Metrics collector
    '''    
    @staticmethod
    def collect(router_entry):
        if router_entry.config_entry.connections:       
            connection_records = IPConnectionDatasource.metric_records(router_entry)
            if connection_records:
                connection_metrics = BaseCollector.gauge_collector('ip_connections_total', 'Number of IP connections', connection_records, 'count',)
                yield connection_metrics

        if router_entry.config_entry.connection_stats:
            connection_stats_records = IPConnectionStatsDatasource.metric_records(router_entry)
            if connection_stats_records:
                for connection_stat_record in connection_stats_records:
                    BaseOutputProcessor.augment_record(router_entry, connection_stat_record, id_key = 'src_address')

                connection_stats_labels = ['src_address', 'dst_addresses', 'dhcp_name']                
                connection_stats_metrics_gauge = BaseCollector.gauge_collector('connection_stats', 'Open connection stats', 
                                                                    connection_stats_records, 'connection_count', connection_stats_labels)
                yield connection_stats_metrics_gauge




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.netwatch_ds import NetwatchMetricsDataSource


class NetwatchCollector(BaseCollector):
    ''' Netwatch Metrics collector
    '''    
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.netwatch:
            return

        netwatch_labels = ['host', 'timeout', 'interval', 'since', 'status', 'comment', 'name', "done_tests", "type", "failed_tests",
                           "loss_count", "loss_percent", "rtt_avg", "rtt_min", "rtt_max", "rtt_jitter", "rtt_stdev",
                           "tcp_connect_time", "http_status_code", "http_resp_time",                           ]
        translation_table = {
            'status': lambda value: '1' if value == 'up' else '0',      
            'rtt_avg': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else '0',
            'rtt_min': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else '0',
            'rtt_max': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else '0',
            'rtt_jitter': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else '0',
            'rtt_stdev': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else '0',
            'tcp_connect_time': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else '0',
            'http_resp_time': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else '0'}
        netwatch_records = NetwatchMetricsDataSource.metric_records(router_entry, metric_labels = netwatch_labels, translation_table=translation_table)

        if netwatch_records:
            netwatch_info_labels = ['host', 'timeout', 'interval', 'since', 'status', 'comment', 'name']
            yield BaseCollector.info_collector('netwatch', 'Netwatch Info Metrics', netwatch_records, netwatch_info_labels)

            yield BaseCollector.gauge_collector('netwatch_status', 'Netwatch Status Metrics', netwatch_records, 'status', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_done_tests', 'Netwatch Done Tests', netwatch_records, 'done_tests', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_failed_tests', 'Netwatch Failed Tests', netwatch_records, 'failed_tests', ['name', 'type'])

            # ICMP specific
            icmp_records = [record for record in netwatch_records if record.get("type", None) == "icmp"]
            yield BaseCollector.gauge_collector('netwatch_icmp_loss_count', 'Netwatch ICMP Loss Count', icmp_records, 'loss_count', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_icmp_loss_percent', 'Netwatch ICMP Loss Percent', icmp_records, 'loss_percent', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_icmp_rtt_avg_ms', 'Netwatch ICMP Round Trip Average', icmp_records, 'rtt_avg', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_icmp_rtt_min_ms', 'Netwatch ICMP Round Trip Min', icmp_records, 'rtt_min', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_icmp_rtt_max_ms', 'Netwatch ICMP Round Trip Max', icmp_records, 'rtt_max', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_icmp_rtt_jitter_ms', 'Netwatch ICMP Round Trip Jitter', icmp_records, 'rtt_jitter', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_icmp_rtt_stdev_ms', 'Netwatch ICMP Round Trip Stdev', icmp_records, 'rtt_stdev', ['name', 'type'])

            # TCP specific
            tcp_records = [record for record in netwatch_records if record.get("type", None) == "tcp-conn"]
            yield BaseCollector.gauge_collector('netwatch_tcp_connect_time_ms', 'Netwatch TCP Connect Time', tcp_records, 'tcp_connect_time', ['name', 'type'])

            # HTTP(s) specific
            http_records = [record for record in netwatch_records if record.get("type", None) in ["http-get", "https-get"]]
            yield BaseCollector.gauge_collector('netwatch_http_status_code', 'Netwatch HTTP status code', http_records, 'http_status_code', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_http_resp_time', 'Netwatch HTTP status code', http_records, 'http_resp_time', ['name', 'type'])
            yield BaseCollector.gauge_collector('netwatch_tcp_connect_time_ms', 'Netwatch TCP Connect Time', http_records, 'tcp_connect_time', ['name', 'type'])



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.bfd_ds import BFDMetricsDataSource


class BFDCollector(BaseCollector):
    """
    Bidirectional Forwarding Detection (BFD) collector
    """
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.bfd:
            return

        translation_table = {
            "actual_tx_interval": lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else "0",
            "desired_tx_interval": lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else "0",
            "hold_time": lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value, ms_span=True) if value else "0",
            "up": lambda value: "1" if value == "true" else "0",
            "uptime": lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value) if value else "0",
        }

        default_labels = ["local_address", "remote_address"]
        metric_records = BFDMetricsDataSource.metric_records(
            router_entry,
            translation_table=translation_table,
        )

        if not metric_records:
            return

        yield BaseCollector.gauge_collector(
            "bfd_multiplier",
            "The multiplier",
            metric_records,
            metric_key="multiplier",
            metric_labels=default_labels
        )
        yield BaseCollector.gauge_collector(
            "bfd_hold_time",
            "The hold time in milliseconds",
            metric_records,
            metric_key="hold_time",
            metric_labels=default_labels
        )
        yield BaseCollector.counter_collector(
            "bfd_rx_packet",
            "BFD control packets received",
            metric_records,
            metric_key="packets_rx",
            metric_labels=default_labels,
        )
        yield BaseCollector.counter_collector(
            "bfd_state_change",
            "Number of time the state changed",
            metric_records,
            metric_key="state_changes",
            metric_labels=default_labels,
        )
        yield BaseCollector.gauge_collector(
            "bfd_tx_interval",
            "The actual transmit interval",
            metric_records,
            metric_key="actual_tx_interval",
            metric_labels=default_labels
        )
        yield BaseCollector.gauge_collector(
            "bfd_tx_interval_desired",
            "Desired transmit interval is the highes value from local tx interval and remote minimum rx interval",
            metric_records,
            metric_key="desired_tx_interval",
            metric_labels=default_labels
        )
        yield BaseCollector.counter_collector(
            "bfd_tx_packet",
            "BFD control packets transmitted",
            metric_records,
            metric_key="packets_tx",
            metric_labels=default_labels,
        )
        yield BaseCollector.gauge_collector(
            "bfd_up",
            "BFD is up",
            metric_records,
            metric_key="up",
            metric_labels=default_labels
        )
        yield BaseCollector.gauge_collector(
            "bfd_uptime",
            "BFD uptime in milliseconds",
            metric_records,
            metric_key="uptime",
            metric_labels=default_labels
        )



#!.usr/bin/env python
# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


import subprocess
from mktxp.cli.config.config import config_handler
from mktxp.cli.options import MKTXPOptionsParser, MKTXPCommands
from mktxp.flow.processor.base_proc import ExportProcessor, OutputProcessor


class MKTXPDispatcher:
    ''' Base MKTXP Commands Dispatcher
    '''
    def __init__(self):
        self.option_parser = MKTXPOptionsParser()

    # Dispatcher
    def dispatch(self):
        args = self.option_parser.parse_options()

        if args['sub_cmd'] == MKTXPCommands.INFO:
            self.print_info()

        elif args['sub_cmd'] == MKTXPCommands.SHOW:
            self.show_entries(args)

        elif args['sub_cmd'] == MKTXPCommands.EXPORT:
            self.start_export(args)

        elif args['sub_cmd'] == MKTXPCommands.PRINT:
            self.print(args)

        elif args['sub_cmd'] == MKTXPCommands.EDIT:
            self.edit_entry(args)

        else:
            # nothing to dispatch
            return False

        return True

    # Dispatched methods
    def print_info(self):
        ''' Prints MKTXP general info
        '''
        print(f'{self.option_parser.script_name}: {self.option_parser.description}')

    def show_entries(self, args):
        if args['config']:
            print(f'MKTXP data config: {config_handler.usr_conf_data_path}')
            print(f'MKTXP internal config: {config_handler.mktxp_conf_path}')
        else:
            for entryname in config_handler.registered_entries():
                if args['entry_name'] and entryname != args['entry_name']:
                    continue
                entry = config_handler.config_entry(entryname)
                print(f'[{entryname}]')
                divider_fields = set(['username', 'use_ssl', 'dhcp'])
                for field in entry._fields:
                    if field == 'password':
                        print(f'    {field}: {"*" * len(entry.password)}')
                    else:
                        if field in divider_fields:
                            print()
                        print(f'    {field}: {getattr(entry, field)}')
                print('\n')

    def edit_entry(self, args):        
        editor = args['editor']
        if not editor:
            print(f'No editor to edit the following file with: {config_handler.usr_conf_data_path}')
        if args['internal']:
            subprocess.check_call([editor, config_handler.mktxp_conf_path])
        else:
            subprocess.check_call([editor, config_handler.usr_conf_data_path])
       
    def start_export(self, args):
        ExportProcessor.start()

    def print(self, args):
        if args['wifi_clients']:
            OutputProcessor.wifi_clients(args['entry_name'])

        elif args['capsman_clients']:
            OutputProcessor.capsman_clients(args['entry_name'])

        elif args['dhcp_clients']:
            OutputProcessor.dhcp_clients(args['entry_name'])

        elif args['conn_stats']:
            OutputProcessor.conn_stats(args['entry_name'])

        else:
            print("Select metric option(s) to print out, or run 'mktxp print -h' to find out more")

def main():
    MKTXPDispatcher().dispatch()

if __name__ == '__main__':
    main()




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.connection_ds import IPConnectionStatsDatasource


class ConnectionsStatsOutput:
    ''' Connections Stats Output
    '''    
    @staticmethod
    def clients_summary(router_entry):
        connection_records = IPConnectionStatsDatasource.metric_records(router_entry, add_router_id = False)
        if not connection_records:
            print('No connection stats records')
            return 

        conn_cnt = 0
        output_records = []
        for registration_record in sorted(connection_records, key = lambda rt_record: rt_record['connection_count'], reverse=True):
            BaseOutputProcessor.resolve_dhcp(router_entry, registration_record, id_key = 'src_address', resolve_address = False)        
            output_records.append(registration_record)
            conn_cnt += registration_record['connection_count']

        output_records_cnt = 0
        output_entry = BaseOutputProcessor.OutputConnStatsEntry
        output_table = BaseOutputProcessor.output_table(output_entry)
        
        for record in output_records:
            output_table.add_row(output_entry(**record))
            output_table.add_row(output_entry())
            output_records_cnt += 1
                
        print (output_table.draw())

        print(f'Distinct source addresses: {output_records_cnt}')
        print(f'Total open connections: {conn_cnt}', '\n')




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.bgp_ds import BGPMetricsDataSource


class BGPCollector(BaseCollector):
    '''BGP collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.bgp:
            return

        bgp_labels = ['name', 'remote_address', 'remote_as', 'local_as', 'remote_afi', 'local_afi', 'remote_messages', 'remote_bytes', 'local_messages', 'local_bytes', 'prefix_count', 'established', 'uptime']
        translation_table = {
                'established': lambda value: '1' if value=='true' else '0',
                'uptime': lambda value: BaseOutputProcessor.parse_timedelta_milliseconds(value) if value else '0'
                }
        bgp_records = BGPMetricsDataSource.metric_records(router_entry, metric_labels=bgp_labels, translation_table = translation_table)
        
        if bgp_records:
            session_info_labes = ['name', 'remote_address', 'remote_as', 'local_as', 'remote_afi', 'local_afi']
            bgp_sessions_metrics = BaseCollector.info_collector('bgp_sessions', 'BGP sessions info', bgp_records, session_info_labes)
            yield bgp_sessions_metrics

            session_id_labes = ['name']
            remote_messages_metrics = BaseCollector.counter_collector('bgp_remote_messages', 'Number of remote messages', bgp_records, 'remote_messages', session_id_labes)
            yield remote_messages_metrics


            local_messages_metrics = BaseCollector.counter_collector('bgp_local_messages', 'Number of local messages', bgp_records, 'local_messages', session_id_labes)
            yield local_messages_metrics


            remote_bytes_metrics = BaseCollector.counter_collector('bgp_remote_bytes', 'Number of remote bytes', bgp_records, 'remote_bytes', session_id_labes)
            yield remote_bytes_metrics


            local_bytes_metrics = BaseCollector.counter_collector('bgp_local_bytes', 'Number of local bytes', bgp_records, 'local_bytes', session_id_labes)
            yield local_bytes_metrics


            prefix_count_metrics = BaseCollector.gauge_collector('bgp_prefix_count', 'BGP prefix count', bgp_records, 'prefix_count', session_id_labes)
            yield prefix_count_metrics


            established_metrics = BaseCollector.gauge_collector('bgp_established', 'BGP established', bgp_records, 'established', session_id_labes)
            yield established_metrics


            uptime_metrics = BaseCollector.gauge_collector('bgp_uptime', 'BGP uptime in milliseconds', bgp_records, 'uptime', session_id_labes)
            yield uptime_metrics




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.cli.config.config import MKTXPConfigKeys
from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.dhcp_ds import DHCPMetricsDataSource


class DHCPCollector(BaseCollector):
    ''' DHCP Metrics collector
    '''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.dhcp:
            return

        dhcp_lease_labels = ['active_address', 'address', 'mac_address', 'host_name', 'comment', 'server', 'expires_after', 'client_id', 'active_mac_address']
        dhcp_lease_records = DHCPMetricsDataSource.metric_records(router_entry, metric_labels = dhcp_lease_labels)
        if dhcp_lease_records:
            # calculate number of leases per DHCP server
            dhcp_lease_servers = {}
            for dhcp_lease_record in dhcp_lease_records:
                if dhcp_lease_record.get('server'):
                    dhcp_lease_servers[dhcp_lease_record['server']] = dhcp_lease_servers.get(dhcp_lease_record['server'], 0) + 1

            # compile leases-per-server records
            dhcp_lease_servers_records = [{ MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                                            MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                                            'server': key, 'count': value} for key, value in dhcp_lease_servers.items()]

            # yield lease-per-server metrics
            dhcp_lease_server_metrics = BaseCollector.gauge_collector('dhcp_lease_active_count',
                                                                        'Number of active leases per DHCP server',
                                                                        dhcp_lease_servers_records, 'count', ['server'])
            yield dhcp_lease_server_metrics

            # active lease metrics
            dhcp_lease_labels.remove('expires_after')
            if router_entry.config_entry.dhcp_lease:
                dhcp_lease_metrics_gauge = BaseCollector.gauge_collector('dhcp_lease_info', 'DHCP Active Leases',
                                                                        dhcp_lease_records, 'expires_after', dhcp_lease_labels)
                yield dhcp_lease_metrics_gauge

            # active lease metrics
            #if router_entry.config_entry.dhcp_lease:
            #    dhcp_lease_metrics = BaseCollector.info_collector('dhcp_lease', 'DHCP Active Leases', dhcp_lease_records, dhcp_lease_labels)
            #    yield dhcp_lease_metrics



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.switch_ds import SwitchPortMetricsDataSource


class SwitchPortCollector(BaseCollector):
    '''Switch Port collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.switch_port:
            return

        switch_port_labels = ['name', 'driver_rx_byte', 'driver_rx_packet', 'driver_tx_byte',  'driver_tx_packet', 
                              'rx_bytes', 'rx_broadcast', 'rx_pause', 'rx_multicast', 'rx_fcs_error', 'rx_align_error', 'rx_fragment', 'rx_overflow', 
                              'tx_bytes', 'tx_broadcast', 'tx_pause', 'tx_multicast', 'tx_underrun', 'tx_collision', 'tx_deferred']

        switch_port_records = SwitchPortMetricsDataSource.metric_records(router_entry, metric_labels = switch_port_labels)
        if switch_port_records:
            for record in switch_port_records:
                for field in switch_port_labels[1:]:
                    if field in record and ',' in record[field]: # https://help.mikrotik.com/docs/display/ROS/Switch+Chip+Features#SwitchChipFeatures-Statistics
                        # Sum each CPU lane for the total
                        record[field] = str(sum([int(count) for count in record[field].split(",")]))
            yield BaseCollector.counter_collector('switch_driver_rx_byte', 'Total count of received bytes', switch_port_records, 'driver_rx_byte', ['name'])
            yield BaseCollector.counter_collector('switch_driver_rx_packet', 'Total count of received packets', switch_port_records, 'driver_rx_packet', ['name'])
            yield BaseCollector.counter_collector('switch_driver_tx_byte', 'Total count of transmitted bytes', switch_port_records, 'driver_tx_byte', ['name'])
            yield BaseCollector.counter_collector('switch_driver_tx_packet', 'Total count of transmitted packets', switch_port_records, 'driver_tx_packet', ['name'])
            yield BaseCollector.counter_collector('switch_rx_bytes', 'Total count of received bytes', switch_port_records, 'rx_bytes', ['name'])
            yield BaseCollector.counter_collector('switch_rx_broadcast', 'Total count of received broadcast frames', switch_port_records, 'rx_broadcast', ['name'])
            yield BaseCollector.counter_collector('switch_rx_pause', 'Total count of received pause frames', switch_port_records, 'rx_pause', ['name'])
            yield BaseCollector.counter_collector('switch_rx_multicast', 'Total count of received multicast frames', switch_port_records, 'rx_multicast', ['name'])
            yield BaseCollector.counter_collector('switch_rx_fcs_error', 'Total count of received frames with incorrect checksum', switch_port_records, 'rx_fcs_error', ['name'])
            yield BaseCollector.counter_collector('switch_rx_align_error', 'Total count of received align error event', switch_port_records, 'rx_align_error', ['name'])
            yield BaseCollector.counter_collector('switch_rx_fragment', 'Total count of received fragmented frames', switch_port_records, 'rx_fragment', ['name'])
            yield BaseCollector.counter_collector('switch_rx_overflow', 'Total count of received overflowed frames', switch_port_records, 'rx_overflow', ['name'])
            yield BaseCollector.counter_collector('switch_tx_bytes', 'Total count of transmitted bytes', switch_port_records, 'tx_bytes', ['name'])
            yield BaseCollector.counter_collector('switch_tx_broadcast', 'Total count of transmitted broadcast frames', switch_port_records, 'tx_broadcast', ['name'])
            yield BaseCollector.counter_collector('switch_tx_pause', 'Total count of transmitted pause frames', switch_port_records, 'tx_pause', ['name'])
            yield BaseCollector.counter_collector('switch_tx_multicast', 'Total count of transmitted multicast frames', switch_port_records, 'tx_multicast', ['name'])
            yield BaseCollector.counter_collector('switch_tx_underrun', 'Total count of transmitted underrun packets', switch_port_records, 'tx_underrun', ['name'])
            yield BaseCollector.counter_collector('switch_tx_collision', 'Total count of transmitted frames that made collisions', switch_port_records, 'tx_collision', ['name'])
            yield BaseCollector.counter_collector('switch_tx_deferred', 'Total count of transmitted frames that were delayed on its first transmit attempt', switch_port_records, 'tx_deferred', ['name'])
